import numpy as np
from genx.core.custom_logging import iprint

_ctype=np.complex128

def refl(kj, kjm1):
    return (kj+kjm1)/2/kjm1

def ass_X_test(k_b, k_u, k_l, dd_u, dd_l,
               sigma, sigma_l, sigma_u, dtype=_ctype):
    ''' Just a simpler version of ass_X to locate an error somewhere...
    '''
    k_jm1_b=k_b[..., 1:]
    k_j_b=k_b[..., :-1]
    k_jm1_l=k_l[..., 1:]
    k_j_u=k_u[..., :-1]
    sigma_jm1_l=sigma_l[..., 1:]
    sigma_j_b=sigma[..., :-1]
    sigma_j_u=sigma_u[..., :-1]
    dd_jm1_l=dd_l[..., 1:]
    dd_j_u=dd_u[..., :-1]

    iprint(dd_j_u, dd_jm1_l)

    # Reflectivites
    r_j_b=refl(k_j_b, k_j_u)
    # print 'r_j_b', r_j_b
    r_jm1_l=refl(k_jm1_l, k_jm1_b)
    # print 'r_jm1_l', r_jm1_l
    r_j_u=refl(k_j_u, k_jm1_l)
    # print 'r_j_u', r_j_u
    p_jm1_l=np.exp(-1.0J*dd_jm1_l*k_jm1_l)
    p_j_u=np.exp(-1.0J*dd_j_u*k_j_u)

    # Defining the X matrix
    X=np.empty((2, 2)+k_j_b.shape, dtype=dtype)

    X[0, 0]=p_jm1_l*r_j_u*p_j_u
    X[0, 1]=-(p_jm1_l*r_j_u-p_jm1_l)/p_j_u
    X[1, 0]=-((r_j_u-1)*p_j_u)/p_jm1_l
    X[1, 1]=r_j_u/(p_jm1_l*p_j_u)

    return X

def ass_X_interfacelayer4(k_b, k_u, k_l, dd_u, dd_l,
                          sigma, sigma_l, sigma_u, dtype=_ctype):
    '''Function that creates a interfacelayer at the top and bottom
    of the layer. with correlated roughness for the interface layers and the
    additional roughness of sigma_l and sigma_u.
    The subscripts b, l and u reffers to bulk, lower and upper respectively.

    v 0.4
    '''
    # print dd_l, dd_u
    # Variables definitions that goes into the expressions:
    k_jm1_b=k_b[..., 1:]
    k_j_b=k_b[..., :-1]
    k_jm1_l=k_l[..., 1:]
    k_j_u=k_u[..., :-1]
    sigma_jm1_l=sigma_l[..., 1:]
    sigma_j_b=sigma[..., :-1]
    sigma_j_u=sigma_u[..., :-1]
    dd_jm1_l=dd_l[..., 1:]
    dd_j_u=dd_u[..., :-1]

    # Reflectivites
    r_j_b=refl(k_j_b, k_j_u)
    r_jm1_l=refl(k_jm1_l, k_jm1_b)
    r_j=refl(k_j_u, k_jm1_l)
    p_jm1_l=np.exp(-1.0J*dd_jm1_l*k_jm1_l)
    p_j_u=np.exp(-1.0J*dd_j_u*k_j_u)
    # p_j_b = np.exp(-1.0J*dd_j_b*k_j_b)

    # Defining the X matrix
    X=np.empty((2, 2)+k_j_b.shape, dtype=dtype)

    tmp1=k_jm1_b**2
    tmp2=sigma_jm1_l**2
    tmp3=-(tmp1*tmp2)/2
    tmp4=-k_jm1_b*tmp2*k_jm1_l
    tmp5=-(tmp2*k_jm1_l**2)/2
    tmp6=sigma_j_b**2
    tmp7=-(tmp1*tmp6)/2
    tmp8=k_j_b**2
    tmp9=-(tmp6*tmp8)/2
    tmp10=sigma_j_u**2
    tmp11=-(tmp8*tmp10)/2
    tmp12=-k_j_b*tmp10*k_j_u
    tmp13=-(tmp10*k_j_u**2)/2
    tmp14=np.exp(tmp13+tmp12+tmp11+tmp9+tmp7+tmp5+tmp4+tmp3)
    tmp15=1/p_jm1_l
    tmp16=1/p_j_u
    tmp17=2*k_jm1_b*tmp2*k_jm1_l
    tmp18=k_jm1_b*tmp6*k_j_b
    tmp21=np.exp(tmp18)
    tmp19=np.exp(tmp17)*tmp21
    tmp20=p_jm1_l**2
    # tmp21=np.exp(tmp18)
    tmp22=-tmp21
    tmp23=2*k_j_b*tmp10*k_j_u
    # tmp24=np.exp(tmp23+tmp18)
    # tmp24=np.exp(tmp23)*tmp21
    # tmp25=np.exp(tmp23+tmp18+tmp17)
    tmp26=p_j_u**2
    # tmp27=np.exp(tmp13+tmp12+tmp11+tmp9-k_jm1_b*tmp6*k_j_b+tmp7+tmp5+tmp4+tmp3)
    tmp27=tmp14/tmp21
    tmp28=np.exp(tmp17)
    # tmp29=np.exp(tmp23+tmp17)
    tmp30=np.exp(tmp23)
    tmp29=tmp28*tmp30
    tmp24=tmp30*tmp21
    tmp25=tmp30*tmp19
    tmp31=-tmp20
    tmp32=tmp21*tmp20
    tmp33=-tmp21*tmp20
    tmp34=tmp24*tmp20

    X[0, 0]=tmp14*tmp15*tmp16*(((((tmp25*tmp20+tmp24)*r_jm1_l-tmp24)*r_j-tmp24*r_jm1_l+tmp24)*tmp26+(
                (tmp19*tmp20+tmp21)*r_jm1_l+tmp22)*r_j-tmp19*tmp20*r_jm1_l)*r_j_b+(
                                           (tmp22-tmp19*tmp20)*r_jm1_l+tmp21)*r_j+tmp19*tmp20*r_jm1_l)

    X[0, 1]=-tmp27*tmp15*tmp16*(((((tmp28*tmp20+1)*r_jm1_l-1)*r_j-r_jm1_l+1)*tmp26+(
                (tmp29*tmp20+tmp30)*r_jm1_l-tmp30)*r_j-tmp29*tmp20*r_jm1_l)*r_j_b+(
                                            ((-tmp28*tmp20-1)*r_jm1_l+1)*r_j+r_jm1_l-1)*tmp26)

    X[1, 0]=-tmp27*tmp15*tmp16*(((((tmp30*tmp20+tmp29)*r_jm1_l-tmp30*tmp20)*r_j-tmp29*r_jm1_l)*tmp26+(
                (tmp20+tmp28)*r_jm1_l+tmp31)*r_j-tmp20*r_jm1_l+tmp20)*r_j_b+(
                                            (tmp31-tmp28)*r_jm1_l+tmp20)*r_j+tmp20*r_jm1_l+tmp31)

    X[1, 1]=tmp14*tmp15*tmp16*(((((tmp32+tmp19)*r_jm1_l+tmp33)*r_j-tmp19*r_jm1_l)*tmp26+(
                (tmp34+tmp25)*r_jm1_l-tmp24*tmp20)*r_j-tmp24*tmp20*r_jm1_l+tmp34)*r_j_b+(
                                           ((tmp33-tmp19)*r_jm1_l+tmp32)*r_j+tmp19*r_jm1_l)*tmp26)

    return X

def ass_X_interfacelayer3(k_b, k_u, k_l, dd_u, dd_l,
                          sigma, sigma_l, sigma_u, dtype=_ctype):
    '''Function that creates a interfacelayer at the top and bottom
    of the layer. with correlated roughness for the interface layers and the
    additional roughness of sigma_l and sigma_u.
    The subscripts b, l and u reffers to bulk, lower and upper respectively.

    v 0.3
    '''

    # Variables definitions that goes into the expressions:
    k_jm1_b=k_b[..., 1:]
    k_j_b=k_b[..., :-1]
    k_jm1_l=k_l[..., 1:]
    k_j_u=k_u[..., :-1]
    sigma_jm1_l=sigma_l[..., 1:]
    sigma_j_b=sigma[..., :-1]
    sigma_j_u=sigma_u[..., :-1]
    dd_jm1_l=dd_l[..., 1:]
    dd_j_u=dd_u[..., :-1]

    iprint(dd_j_u, dd_jm1_l)

    # Reflectivites
    r_j_b=refl(k_j_b, k_j_u)
    iprint('r_j_b', r_j_b)
    r_jm1_l=refl(k_jm1_l, k_jm1_b)
    iprint('r_jm1_l', r_jm1_l)
    r_j_u=refl(k_j_u, k_jm1_l)
    iprint('r_j_u', r_j_u)
    p_jm1_l=np.exp(-1.0J*dd_jm1_l*k_jm1_l)
    p_j_u=np.exp(-1.0J*dd_j_u*k_j_u)
    # p_j_b = np.exp(-1.0J*dd_j_b*k_j_b)

    # Defining the X matrix
    X=np.empty((2, 2)+k_j_b.shape, dtype=dtype)

    tmp1=k_jm1_b**2
    tmp2=sigma_jm1_l**2
    tmp3=-(tmp1*tmp2)/2
    tmp4=k_jm1_b*tmp2*k_jm1_l
    tmp5=-(tmp2*k_jm1_l**2)/2
    tmp6=sigma_j_b**2
    tmp7=-(tmp1*tmp6)/2
    tmp8=k_jm1_b*tmp6*k_j_b
    tmp9=k_j_b**2
    tmp10=-(tmp6*tmp9)/2
    tmp11=sigma_j_u**2
    tmp12=-(tmp9*tmp11)/2
    tmp13=-k_j_b*tmp11*k_j_u
    tmp14=-(tmp11*k_j_u**2)/2
    tmp15=np.exp(tmp14+tmp13+tmp12+tmp10+tmp8+tmp7+tmp5+tmp4+tmp3)
    tmp16=1/p_j_u
    tmp17=-k_jm1_b*tmp2*k_jm1_l
    tmp18=np.exp(tmp14+tmp13+tmp12+tmp10+tmp8+tmp7+tmp5+tmp17+tmp3)
    tmp19=1/p_jm1_l
    tmp20=k_j_b*tmp11*k_j_u
    tmp21=np.exp(tmp14+tmp20+tmp12+tmp10+tmp8+tmp7+tmp5+tmp17+tmp3)
    tmp22=np.exp(tmp14+tmp20+tmp12+tmp10+tmp8+tmp7+tmp5+tmp4+tmp3)
    tmp23=-k_jm1_b*tmp6*k_j_b
    tmp24=np.exp(tmp14+tmp13+tmp12+tmp10+tmp23+tmp7+tmp5+tmp17+tmp3)
    tmp25=np.exp(tmp14+tmp13+tmp12+tmp10+tmp23+tmp7+tmp5+tmp4+tmp3)
    tmp26=np.exp(tmp14+tmp20+tmp12+tmp10+tmp23+tmp7+tmp5+tmp4+tmp3)
    tmp27=np.exp(tmp14+tmp20+tmp12+tmp10+tmp23+tmp7+tmp5+tmp17+tmp3)

    X[
        0, 0]=tmp22*p_jm1_l*r_jm1_l*r_j_b*p_j_u*r_j_u+tmp21*tmp19*r_jm1_l*r_j_b*p_j_u*r_j_u-tmp21*tmp19*r_j_b*p_j_u*r_j_u-tmp21*tmp19*r_jm1_l*p_j_u*r_j_u+tmp21*tmp19*p_j_u*r_j_u+tmp15*p_jm1_l*r_jm1_l*r_j_b*tmp16*r_j_u+tmp18*tmp19*r_jm1_l*r_j_b*tmp16*r_j_u-tmp18*tmp19*r_j_b*tmp16*r_j_u-tmp15*p_jm1_l*r_jm1_l*tmp16*r_j_u-tmp15*p_jm1_l*r_jm1_l*r_j_b*tmp16-tmp18*tmp19*r_jm1_l*r_j_b*tmp16+tmp18*tmp19*r_j_b*tmp16+tmp15*p_jm1_l*r_jm1_l*tmp16

    X[
        0, 1]=-tmp25*p_jm1_l*r_jm1_l*r_j_b*p_j_u*r_j_u-tmp24*tmp19*r_jm1_l*r_j_b*p_j_u*r_j_u+tmp24*tmp19*r_j_b*p_j_u*r_j_u+tmp24*tmp19*r_jm1_l*p_j_u*r_j_u-tmp24*tmp19*p_j_u*r_j_u-tmp26*p_jm1_l*r_jm1_l*r_j_b*tmp16*r_j_u-tmp27*tmp19*r_jm1_l*r_j_b*tmp16*r_j_u+tmp27*tmp19*r_j_b*tmp16*r_j_u+tmp26*p_jm1_l*r_jm1_l*tmp16*r_j_u+tmp25*p_jm1_l*r_jm1_l*r_j_b*p_j_u+tmp24*tmp19*r_jm1_l*r_j_b*p_j_u-tmp24*tmp19*r_j_b*p_j_u-tmp24*tmp19*r_jm1_l*p_j_u+tmp24*tmp19*p_j_u

    X[
        1, 0]=-tmp27*p_jm1_l*r_jm1_l*r_j_b*p_j_u*r_j_u-tmp26*tmp19*r_jm1_l*r_j_b*p_j_u*r_j_u+tmp27*p_jm1_l*r_j_b*p_j_u*r_j_u+tmp26*tmp19*r_jm1_l*p_j_u*r_j_u-tmp24*p_jm1_l*r_jm1_l*r_j_b*tmp16*r_j_u-tmp25*tmp19*r_jm1_l*r_j_b*tmp16*r_j_u+tmp24*p_jm1_l*r_j_b*tmp16*r_j_u+tmp24*p_jm1_l*r_jm1_l*tmp16*r_j_u-tmp24*p_jm1_l*tmp16*r_j_u+tmp24*p_jm1_l*r_jm1_l*r_j_b*tmp16+tmp25*tmp19*r_jm1_l*r_j_b*tmp16-tmp24*p_jm1_l*r_j_b*tmp16-tmp24*p_jm1_l*r_jm1_l*tmp16+tmp24*p_jm1_l*tmp16

    X[
        1, 1]=tmp18*p_jm1_l*r_jm1_l*r_j_b*p_j_u*r_j_u+tmp15*tmp19*r_jm1_l*r_j_b*p_j_u*r_j_u-tmp18*p_jm1_l*r_j_b*p_j_u*r_j_u-tmp15*tmp19*r_jm1_l*p_j_u*r_j_u+tmp21*p_jm1_l*r_jm1_l*r_j_b*tmp16*r_j_u+tmp22*tmp19*r_jm1_l*r_j_b*tmp16*r_j_u-tmp21*p_jm1_l*r_j_b*tmp16*r_j_u-tmp21*p_jm1_l*r_jm1_l*tmp16*r_j_u+tmp21*p_jm1_l*tmp16*r_j_u-tmp18*p_jm1_l*r_jm1_l*r_j_b*p_j_u-tmp15*tmp19*r_jm1_l*r_j_b*p_j_u+tmp18*p_jm1_l*r_j_b*p_j_u+tmp15*tmp19*r_jm1_l*p_j_u

    return X

def ass_X_interfacelayer(k_b, k_u, k_l, dd_u, dd_l,
                         sigma, sigma_l, sigma_u, dtype=_ctype):
    '''Function that creates a interfacelayer at the top and bottom
    of the layer. with correlated roughness for the interface layers and the
    additional roughness of sigma_l and sigma_u.
    The subscripts b, l and u reffers to bulk, lower and upper respectively.

    v 0.2
    '''
    iprint(dd_l, dd_u)
    # Variables definitions that goes into the expressions:
    k_jm1_b=k_b[..., 1:]
    k_j_b=k_b[..., :-1]
    k_jm1_l=k_l[..., 1:]
    k_j_u=k_u[..., :-1]
    sigma_jm1_l=sigma_l[..., 1:]
    sigma_j_b=sigma[..., :-1]
    sigma_j_u=sigma_u[..., :-1]
    dd_jm1_l=dd_l[..., 1:]
    dd_j_u=dd_u[..., :-1]

    # Reflectivites
    r_j_b=refl(k_j_b, k_j_u)
    r_jm1_l=refl(k_jm1_l, k_jm1_b)
    r_j_u=refl(k_j_u, k_jm1_l)
    p_jm1_l=np.exp(-1.0J*dd_jm1_l*k_jm1_b)
    p_j_u=np.exp(-1.0J*dd_j_u*k_j_u)
    # p_j_b = np.exp(-1.0J*dd_j_b*k_j_b)

    # Defining the X matrix
    X=np.empty((2, 2)+k_j_b.shape, dtype=dtype)

    tmp1=k_jm1_b**2
    tmp2=sigma_jm1_l**2
    tmp3=-(tmp1*tmp2)/2
    tmp4=-k_jm1_b*tmp2*k_jm1_l
    tmp5=-(tmp2*k_jm1_l**2)/2
    tmp6=sigma_j_b**2
    tmp7=-(tmp1*tmp6)/2
    tmp8=k_j_b**2
    tmp9=-(tmp6*tmp8)/2
    tmp10=sigma_j_u**2
    tmp11=-(tmp8*tmp10)/2
    tmp12=-k_j_b*tmp10*k_j_u
    tmp13=-(tmp10*k_j_u**2)/2
    tmp14=np.exp(tmp13+tmp12+tmp11+tmp9+tmp7+tmp5+tmp4+tmp3)
    tmp15=1/p_jm1_l
    tmp16=1/p_j_u
    tmp17=2*k_jm1_b*tmp2*k_jm1_l
    tmp18=k_jm1_b*tmp6*k_j_b
    tmp21=np.exp(tmp18)
    tmp19=np.exp(tmp17)*tmp21
    tmp20=p_jm1_l**2
    # tmp21=np.exp(tmp18)
    tmp22=-tmp21
    tmp23=2*k_j_b*tmp10*k_j_u
    # tmp24=np.exp(tmp23+tmp18)
    # tmp24=np.exp(tmp23)*tmp21
    # tmp25=np.exp(tmp23+tmp18+tmp17)
    tmp26=p_j_u**2
    # tmp27=np.exp(tmp13+tmp12+tmp11+tmp9-k_jm1_b*tmp6*k_j_b+tmp7+tmp5+tmp4+tmp3)
    tmp27=tmp14/tmp21
    tmp28=np.exp(tmp17)
    # tmp29=np.exp(tmp23+tmp17)
    tmp30=np.exp(tmp23)
    tmp29=tmp28*tmp30
    tmp24=tmp30*tmp21
    tmp25=tmp30*tmp19
    tmp31=-tmp20
    tmp32=tmp21*tmp20
    tmp33=-tmp21*tmp20
    tmp34=tmp24*tmp20

    X[0, 0]=tmp14*tmp15*tmp16*(((((tmp25*tmp20+tmp24)*r_jm1_l-tmp24)*r_j_b-tmp24*r_jm1_l+tmp24)*tmp26+(
                (tmp19*tmp20+tmp21)*r_jm1_l+tmp22)*r_j_b-tmp19*tmp20*r_jm1_l)*r_j_u+(
                                           (tmp22-tmp19*tmp20)*r_jm1_l+tmp21)*r_j_b+tmp19*tmp20*r_jm1_l)

    X[0, 1]=-tmp27*tmp15*tmp16*(((((tmp28*tmp20+1)*r_jm1_l-1)*r_j_b-r_jm1_l+1)*tmp26+(
                (tmp29*tmp20+tmp30)*r_jm1_l-tmp30)*r_j_b-tmp29*tmp20*r_jm1_l)*r_j_u+(
                                            ((-tmp28*tmp20-1)*r_jm1_l+1)*r_j_b+r_jm1_l-1)*tmp26)

    X[1, 0]=-tmp27*tmp15*tmp16*(((((tmp30*tmp20+tmp29)*r_jm1_l-tmp30*tmp20)*r_j_b-tmp29*r_jm1_l)*tmp26+(
                (tmp20+tmp28)*r_jm1_l+tmp31)*r_j_b-tmp20*r_jm1_l+tmp20)*r_j_u+(
                                            (tmp31-tmp28)*r_jm1_l+tmp20)*r_j_b+tmp20*r_jm1_l+tmp31)

    X[1, 1]=tmp14*tmp15*tmp16*(((((tmp32+tmp19)*r_jm1_l+tmp33)*r_j_b-tmp19*r_jm1_l)*tmp26+(
                (tmp34+tmp25)*r_jm1_l-tmp24*tmp20)*r_j_b-tmp24*tmp20*r_jm1_l+tmp34)*r_j_u+(
                                           ((tmp33-tmp19)*r_jm1_l+tmp32)*r_j_b+tmp19*r_jm1_l)*tmp26)

    return X

def ass_X_interfacelayer1(k_b, k_u, k_l, dd_u, dd_l,
                          sigma, sigma_l, sigma_u, dtype=_ctype):
    '''Function that creates a interfacelayer at the top and bottom
    of the layer. with correlated roughness for the interface layers and the
    additional roughness of sigma_l and sigma_u.
    The subscripts b, l and u reffers to bulk, lower and upper respectively.

    v 0.1
    '''

    # Variables definitions that goes into the expressions:
    k_jm1_b=k_b[..., 1:]
    k_j_b=k_b[..., :-1]
    k_jm1_l=k_l[..., 1:]
    k_j_u=k_u[..., :-1]
    sigma_jm1_l=sigma_l[..., 1:]
    sigma_j_b=sigma[..., :-1]
    sigma_j_u=sigma_u[..., :-1]
    dd_jm1_l=dd_u[..., 1:]
    dd_j_u=dd_u[..., :-1]

    # Defining the X matrix
    X=np.empty((2, 2)+k_j_b.shape, dtype=dtype)

    # Temporary variables to ease the calculation
    tmp1=k_jm1_b*k_jm1_b
    tmp2=sigma_jm1_l*sigma_jm1_l
    tmp3=tmp1*tmp2
    tmp4=(2*k_jm1_b*tmp2+2*1.0J*dd_jm1_l)*k_jm1_l
    tmp5=k_jm1_l*k_jm1_l
    tmp6=tmp2*tmp5
    tmp7=sigma_j_b*sigma_j_b
    tmp8=tmp1*tmp7
    tmp9=k_j_b**2
    tmp10=tmp7*tmp9
    tmp11=sigma_j_u*sigma_j_u
    tmp12=tmp9*tmp11
    tmp13=(2*k_j_b*tmp11+2*1.0J*dd_j_u)*k_j_u
    tmp14=k_j_u*k_j_u
    tmp15=tmp11*tmp14
    tmp16=1/np.exp((tmp15+tmp13+tmp12+tmp10+tmp8+tmp6+tmp4+tmp3)/2)
    tmp17=1/k_jm1_b
    tmp18=1/k_jm1_l
    tmp19=1/k_j_u
    tmp20=np.exp(2*1.0J*dd_jm1_l*k_jm1_l)
    tmp21=-tmp20
    tmp22=np.exp(2*k_jm1_b*tmp2*k_jm1_l)
    tmp23=-tmp22
    tmp24=tmp23+tmp21
    tmp25=k_jm1_b*tmp7*k_j_b
    tmp26=2*1.0J*dd_j_u*k_j_u
    tmp27=np.exp(tmp26+tmp25)
    tmp28=tmp22+tmp20
    tmp29=2*k_j_b*tmp11*k_j_u
    tmp30=np.exp(tmp29+tmp25)
    tmp31=tmp28*tmp30
    tmp32=tmp31+tmp24*tmp27
    tmp33=tmp23+tmp20
    tmp34=tmp22+tmp21
    tmp35=tmp34*tmp30
    tmp36=tmp35+tmp33*tmp27
    tmp37=tmp31+tmp28*tmp27
    tmp38=tmp35+tmp34*tmp27
    tmp39=1/np.exp((tmp15+tmp13+tmp12+tmp10+2*k_jm1_b*tmp7*k_j_b+tmp8+tmp6+tmp4+tmp3)/2)
    tmp40=np.exp(tmp13)
    tmp41=tmp24*tmp40
    tmp42=tmp33*tmp40
    tmp43=tmp34*tmp40
    tmp44=tmp28*tmp40
    tmp45=np.exp(tmp4)
    tmp46=tmp45+1
    tmp47=np.exp(tmp26)
    tmp48=tmp46*tmp47
    tmp49=-tmp45
    tmp50=tmp49-1
    tmp51=np.exp(tmp29)
    tmp52=tmp50*tmp51
    tmp53=tmp45-1
    tmp54=tmp53*tmp47
    tmp55=tmp49+1
    tmp56=tmp55*tmp51
    tmp57=tmp50*tmp47
    tmp58=tmp55*tmp47
    tmp59=tmp53*tmp51
    tmp60=tmp46*tmp51
    tmp61=np.exp(tmp25)
    tmp62=np.exp(tmp13+tmp25)
    tmp63=tmp46*tmp62
    tmp64=tmp63+tmp50*tmp61
    tmp65=tmp53*tmp62
    tmp66=tmp65+tmp55*tmp61
    tmp67=tmp63+tmp46*tmp61
    tmp68=tmp65+tmp53*tmp61

    # Assembling the interface matrices
    X[0, 0]=(tmp16*tmp17*tmp18*tmp19*((tmp32*k_jm1_l+tmp36*k_jm1_b)*tmp14+(
                (tmp37*k_jm1_l+tmp38*k_jm1_b)*k_j_b+tmp38*tmp5+tmp37*k_jm1_b*k_jm1_l)*k_j_u+(
                                                  tmp36*tmp5+tmp32*k_jm1_b*k_jm1_l)*k_j_b))/8

    X[0, 1]=-(tmp39*tmp17*tmp18*tmp19*(((tmp44+tmp23+tmp21)*k_jm1_l+(tmp43+tmp23+tmp20)*k_jm1_b)*tmp14+(
                ((tmp44+tmp22+tmp20)*k_jm1_l+(tmp43+tmp22+tmp21)*k_jm1_b)*k_j_b+(tmp42+tmp23+tmp20)*tmp5+(
                    tmp41+tmp23+tmp21)*k_jm1_b*k_jm1_l)*k_j_u+((tmp42+tmp22+tmp21)*tmp5+(
                tmp41+tmp22+tmp20)*k_jm1_b*k_jm1_l)*k_j_b))/8

    X[1, 0]=-(tmp39*tmp17*tmp18*tmp19*(((tmp60+tmp57)*k_jm1_l+(tmp59+tmp58)*k_jm1_b)*tmp14+(
                ((tmp60+tmp48)*k_jm1_l+(tmp59+tmp54)*k_jm1_b)*k_j_b+(tmp56+tmp58)*tmp5+(
                    tmp52+tmp57)*k_jm1_b*k_jm1_l)*k_j_u+((tmp56+tmp54)*tmp5+(tmp52+tmp48)*k_jm1_b*k_jm1_l)*k_j_b))/8

    X[1, 1]=(tmp16*tmp17*tmp18*tmp19*((tmp64*k_jm1_l+tmp66*k_jm1_b)*tmp14+(
                (tmp67*k_jm1_l+tmp68*k_jm1_b)*k_j_b+tmp68*tmp5+tmp67*k_jm1_b*k_jm1_l)*k_j_u+(
                                                  tmp66*tmp5+tmp64*k_jm1_b*k_jm1_l)*k_j_b))/8

    return X



"""
A remote client that performs model refinement on a mashine running genx.server.

Currently, only differential evolution is supported.
"""

import asyncio
import struct
from threading import Thread
from dataclasses import dataclass
from hashlib import blake2b
from logging import debug

from numpy import array

from . import messaging
from ..core import AUTH_SIZE, HANDSHAKE1, HANDSHAKE2
from ..core.config import BaseConfig
from ..model import Model
from ..solver_basis import GenxOptimizer, GenxOptimizerCallback, SolverParameterInfo, SolverResultInfo, SolverUpdateInfo
from ..diffev import DiffEvConfig, DiffEvDefaultCallbacks


@dataclass
class RemoteDiffEvConfig(BaseConfig):
    section = 'solver'

    address: str = 'localhost'
    port: int = 3000
    key: str = 'empty'

    km: float = BaseConfig.GParam(0.7, pmin=0.0, pmax=1.0)
    kr: float = BaseConfig.GParam(0.7, pmin=0.0, pmax=1.0)
    allowed_fom_discrepancy: float = 1e-10

    use_pop_mult: bool = True
    pop_size: int = BaseConfig.GParam(50, pmin=5, pmax=10000, label='Fixed size')
    pop_mult: int = BaseConfig.GParam(3, pmin=1, pmax=100, label='Relative size')
    create_trial: str = BaseConfig.GChoice('best_1_bin', ['best_1_bin', 'rand_1_bin', 'best_either_or',
                                                          'rand_either_or', 'jade_best', 'simplex_best_1_bin'],
                                           label='Method')

    use_max_generations: bool = False
    max_generations: int = BaseConfig.GParam(500, pmin=10, pmax=10000, label='Fixed size')
    max_generation_mult: int = BaseConfig.GParam(6, pmin=1, pmax=100, label='Relative size')
    min_parameter_spread: float = BaseConfig.GParam(0.0, pmin=0.0, pmax=100.0, label='parameter spread to stop (%)')

    use_start_guess: bool = True
    use_boundaries: bool = True

    max_log_elements: int = BaseConfig.GParam(100000, pmin=1000, pmax=1000000, label=', # elements')
    use_mpi: bool = False
    use_parallel_processing: bool = False
    parallel_processes: int = BaseConfig.GParam(16, pmin=2, pmax=1000, label='# processes')
    parallel_chunksize: int = BaseConfig.GParam(10, pmin=1, pmax=1000, label='items/chunk')

    use_autosave: bool = False
    autosave_interval: int = BaseConfig.GParam(10, pmin=1, pmax=1000, label=', interval')

    save_all_evals: bool = False
    errorbar_level: float = BaseConfig.GParam(1.05, pmin=1.001, pmax=2.0)

    groups = {  # for building config dialogs
        'Server': ['address', 'port', 'key'],
        'Fitting': [['use_start_guess', 'use_boundaries'], ['use_autosave', 'autosave_interval'],
                    ['save_all_evals', 'max_log_elements']],
        'Differential Evolution':
            ['km', 'kr', 'create_trial',
             ['Population size:', 'use_pop_mult', 'pop_mult', 'pop_size'],
             ['Max. Generations:', 'use_max_generations', 'max_generations', 'max_generation_mult'],
             'min_parameter_spread',
             ],
        'Parallel processing': ['use_mpi', 'use_parallel_processing', 'parallel_processes', 'parallel_chunksize']
        }


class RemoteOptimizer(GenxOptimizer):
    opt: RemoteDiffEvConfig
    n_fom_evals = 0

    _callbacks: GenxOptimizerCallback = DiffEvDefaultCallbacks()

    def __init__(self):
        GenxOptimizer.__init__(self)

        # Control flags:
        self.running = False  # true if optimization is running
        self.stop = False  # true if the optimization should stop
        self.setup_ok = False  # True if the optimization have been setup
        self.error = None  # None/string if an error ahs occurred

        # Logging variables
        self.fom_log = array([[0, 0]])[0:0]

        self.start_guess = array([])
        self.updated_kr = []
        self.updated_km = []

        self.reader = None
        self.writer = None

    def pickle_string(self, clear_evals: bool = False):
        pass

    def pickle_load(self, pickled_string: bytes):
        pass

    def is_running(self) -> bool:
        return self.running

    def get_start_guess(self):
        return self.start_guess

    def get_model(self) -> Model:
        pass

    def get_fom_log(self):
        pass

    def calc_error_bar(self, index: int) -> (float, float):
        pass

    def project_evals(self, index: int):
        pass

    def start_fit(self, model: Model):
        '''
        Starts fitting on a remote server.
        '''
        if not self.running:
            debug('Start asyncio task for fit')
            self.n_fom_evals = 0
            self.start_guess = array([])
            thread=Thread(target=self._start_remote_fit, args=(model,))
            thread.start()
            return True
        else:
            self.text_output('Fit is already running, stop and then start')
            return False

    def model_message(self, model: Model) -> messaging.ModelTransfer:
        opt = DiffEvConfig()
        for key in opt.asdict():
            # transfer all options to send config
            setattr(opt, key, getattr(self.opt, key))
        mm = messaging.ModelTransfer(model, opt)
        return mm

    def _start_remote_fit(self, model: Model):
        asyncio.run(self.start_remote_fit(model))

    async def start_remote_fit(self, model: Model):
        if not self.running:
            self.stop = False
            self.text_output('Trying to connect to server...')
            await self.connect()
            debug('Connection successful')
            await self.send_message(self.model_message(model))
            self.text_output('Starting the fit...')
            await self.send_message(messaging.ActionMessage(messaging.ActionType.START_FIT, 'Start fit', ''))
            self.running = True
            await self.receive()
        else:
            raise RuntimeError('Trying to start new fit while remote fit should be running, '
                               'this is unexpected behavior')

    async def connect(self):
        self.reader, self.writer = await asyncio.open_connection(self.opt.address, self.opt.port)
        key = self.opt.key.encode('utf-8')
        ref1 = blake2b(HANDSHAKE1, key=key, digest_size=AUTH_SIZE).hexdigest().encode('ascii')
        ref2 = blake2b(HANDSHAKE2, key=key, digest_size=AUTH_SIZE).hexdigest().encode('ascii')
        debug(f'Sending handshake message to {self.opt.address}:{self.opt.port}')
        self.writer.write(ref1)
        await self.writer.drain()
        debug('Receiving response message')
        fb = await self.reader.read(len(ref2))
        if ref2!=fb:
            debug('Response wrong, cancel connection')
            self.writer.close()
            await self.writer.wait_closed()
            self.reader = None
            self.writer = None

    async def send_message(self, message: messaging.GenXMessage):
        debug(f'send_message {message}')
        self.writer.write(message.message())
        await self.writer.drain()

    async def receive(self):
        while self.running:
            try:
                obj = await messaging.GenXMessage.receive(self.reader)
            except struct.error:
                debug('strcut.error raised, connection interrupted')
                self.running = False
                return
            debug(f'remote send message: {obj!r}')
            if isinstance(obj, messaging.StingMessage):
                self._callbacks.text_output(obj.text)
            elif isinstance(obj, messaging.OptimizerUpdate):
                if isinstance(obj.payload, SolverUpdateInfo):
                    self._callbacks.plot_output(obj.payload)
                elif isinstance(obj.payload, SolverParameterInfo):
                    if isinstance(obj.payload, SolverResultInfo):
                        debug('Fit ended on remote')
                        self.running = False
                        self.new_best = obj.payload.start_guess
                        self.start_guess = obj.payload.start_guess
                        self.n_fom_evals = len(obj.payload.population)
                        self._callbacks.fitting_ended(obj.payload)
                    else:
                        self._callbacks.parameter_output(obj.payload)
            else:
                raise NotImplemented(f'Unable to handle message; {obj!r}')

    def stop_fit(self):
        self.text_output('trying to stop the fit')
        asyncio.run(self.send_message(messaging.ActionMessage(messaging.ActionType.STOP_FIT, 'Stop fit', '')))

    def resume_fit(self, model: Model):
        pass

    def is_fitted(self):
        pass

    def is_configured(self):
        pass

    def set_callbacks(self, callbacks: GenxOptimizerCallback):
        self._callbacks = callbacks

    def get_callbacks(self) -> GenxOptimizerCallback:
        return self._callbacks

    def get_result_info(self) -> SolverResultInfo:
        pass

    def text_output(self, text: str):
        self._callbacks.text_output(text)



''' utils.py

module for utils needed for the plugin frameworks. This module implemtents a
handler which searches for plugins and loads and removes them.
Programmer: Matts Bjorck
Last changed: 2008 07 23
'''

import os
import sys
from logging import info, warning, error

try:
    import wx
except ImportError:
    wx=None

# ==============================================================================

class PluginHandler:
    ''' Class that takes care of the loading/unloading of the plugins.
    This is used in the main GUI. Should not be tinkered with. 
    The plugin module can only conatain files wich are real plugins i.e.
    that implements the class Template. Of course it is possible to have
    subdirectories that has support libraries.
    
    Note that the plugins modules has to have an class called Plugin
    and that this class should implement the method Remove. The init function
    should only take the parent (i.e. the original window frame) as input.
    '''

    def __init__(self, parent, directory, plugin_folder):
        '''__init__(self, parent)
        Create an instance of PluginHandler
        '''
        self.parent=parent
        # self.modules = []
        self.loaded_plugins={}
        self.directory=directory
        self.plugin_folder=plugin_folder

    def get_plugins(self):
        '''get_possible_plugins(self) --> list of strings
        
        search the plugin directory. 
        Checks the list for python scripts and returns a list of 
        module names that are NOT loaded.
        '''
        # Locate all python files in this files directory
        # but excluding this file and not loaded.
        plugins=[s.rsplit('.', 1)[0] for s in os.listdir(self.directory \
                                                         +self.plugin_folder) if
                 s.split('.', 1)[-1] in ['py', 'pyo', 'pyc']
                 and s[:2]!='__' and \
                 s.rsplit('.', 1)[0] not in self.loaded_plugins]
        return plugins

    def get_possible_plugins(self):
        '''get_possible_plugins(self) --> list of strings
        
        search the plugin directory. 
        Checks the list for python scripts and returns a list of 
        module names that are loadable .
        '''
        # Locate all python files in this files directory
        # but excluding this file and not loaded.
        plugins=[s[:-3] for s in os.listdir(self.directory \
                                            +self.plugin_folder) if '.py'==s[-3:]
                 and s[:2]!='__']
        return plugins

    def get_loaded_plugins(self):
        '''get_loaded_plugins(self) --> plugins [list]
        
        returns a list of the loaded plugins
        '''
        return list(self.loaded_plugins.keys())

    def load_plugin(self, plugin_name):
        '''load_plugin(self, plugin_name) --> plugin object
        load the plugin given by the plugin_name [string].
        '''
        # Load the module
        module=self._load_module(self.plugin_folder+'.'+plugin_name)
        # Create the Plugin object from the module
        plugin=module.Plugin(self.parent)
        # Add it to our dictonary over loaded plugins
        self.loaded_plugins[plugin_name]=plugin

    def is_loaded(self, plugin_name):
        '''isloaded(self, plugin_name) --> loaded [bool]
        
        Returns True if plugin with name plugin_name is loaded, 
        otherwise False.
        '''
        return plugin_name in self.loaded_plugins

    def _load_module(self, module_name):
        ''' _load_module(self, module) --> module
        Load a module given by name
        '''
        # print 'Trying to load module: ', module_name
        module=__import__(module_name, globals(), locals(), ['plugins'], level=1)
        return module

    def unload_plugin(self, plugin_name):
        ''' unload_plugin(self, plugin_name) --> None
        Used to remove the plugin from the system.
        '''
        # print self.loaded_plugins.keys()
        self.loaded_plugins[plugin_name].Remove()
        del self.loaded_plugins[plugin_name]

# END: PluginHandler
# ==============================================================================
# Utility Dialog functions..
def ShowInfoDialog(frame, message, title='Information'):
    if wx is None:
        print(message)
        return
    else:
        exc_info=sys.exc_info()
        info(message, exc_info=exc_info[0] and exc_info)
    dlg=wx.MessageDialog(frame, message, title,
                         wx.OK | wx.ICON_INFORMATION
                         )
    dlg.ShowModal()
    dlg.Destroy()

def ShowWarningDialog(frame, message, title='Warning'):
    if wx is None:
        print(message)
        return
    else:
        exc_info=sys.exc_info()
        warning(message, exc_info=exc_info[0] and exc_info)
    dlg=wx.MessageDialog(frame, message, title,
                         wx.OK | wx.ICON_ERROR
                         )
    dlg.ShowModal()
    dlg.Destroy()

def ShowErrorDialog(frame, message, title='ERROR'):
    if wx is None:
        print(message)
        return
    else:
        exc_info=sys.exc_info()
        error(message, exc_info=exc_info[0] and exc_info)
    dlg=wx.MessageDialog(frame, message, title,
                         wx.OK | wx.ICON_ERROR
                         )
    dlg.ShowModal()
    dlg.Destroy()

def ShowQuestionDialog(frame, message, title='Question'):
    if wx is None:
        result = input(message+' [y/n]').strip().lower() in ['y', 'yes']
        return result
    dlg=wx.MessageDialog(frame, message,
                         title,
                         wx.YES_NO | wx.ICON_QUESTION
                         )
    result=dlg.ShowModal()==wx.ID_YES
    dlg.Destroy()
    return result



''' <h1>Reflectivity plugin </h1>
Reflectivity is a plugin for providing a graphical user
interface to define multilayer structures in GenX. It works
on quite general principels with dynamic generation of the
graphical user interface depending on the model. It also
dynamically generates the script for the model. Thus, it is 
always possible to go in and edit the script manually. <p>

The plugin consists of the following components:
<h2>Sample tab</h2>
This tab has the definitons for the layers and stacks. 
remember that a layer has to be inside a stack. 
Also, the name of the layers must be uniqe and can not be change
after the layer has been created. The functions of the buttons 
from left to right are:
<dl>
    <dt><b>Add Layer</b></dt>
    <dd>Add a new layer to the current position</dd>
    <dt><b>Add Stack</b></dt>
    <dd>Add a new stack to the current position</dd>
    <dt><b>Remove item</b></dt>
    <dd>Removes the current item. Note that the substrate, Sub, and the 
    ambient material can not be removed.</dd>
    <dt><b>Move item up</b></dt>
    <dd>Move item up</dd>
    <dt><b>Move item down</b></dt>
    <dd>Move item down</dd>
    <dt><b>Sample parameters</b></dt>
    <dd>Edit global parameters for the entire sample</dd>
    <dt><b>Instrument</b></dt>
    <dd>Edit parameters such as resolution and incident intesnity that
    is defines the instrument.</dd>
</dl>
<h2>Simulation tab</h2>
Here it is possible to add commands that are conducted before a data 
set is calculated. This done by adding a new command by pressing the green
plus sign. This brings up a dialog where the object and paraemter can
be chosen and the expression typed in. Note that the list one can choose from 
is <b>only</b> updated when the simulation button is pressed.<p>

The blue nut button to the right brings up a menu that allows the definition
of custom variables. These can be used to define problem specific parameters
 such as, for example, compostion of layers. One can also use this for parameter
coupling that yields a speedup in fitting. For example, fitting the repetition
length for a multilayer. 

<h2>SLD tab</h2>
This shows the real and imaginary part of the scattering length as a function
of depth for the sample. The substrate is to the left and the ambient material
is to the right. This is updated when the simulation button is pressed.
'''
import io, traceback
import wx.html

from .help_modules.reflectivity_misc import ReflectivityModule
from .help_modules.reflectivity_sample_plot import SamplePlotPanel
from .help_modules.custom_dialog import *
from .help_modules.reflectivity_utils import SampleHandler, SampleBuilder, avail_models
from .help_modules.reflectivity_gui import SamplePanel, DataParameterPanel
from .. import add_on_framework as framework
from genx.exceptions import GenxError
from genx.core.custom_logging import iprint
from genx.gui.custom_events import EVT_UPDATE_SCRIPT
from genx.model import Model


class Plugin(framework.Template, SampleBuilder, wx.EvtHandler):
    previous_xaxis = None
    _last_script = None
    model: ReflectivityModule
    sample_widget: SamplePanel
    sampleh: SampleHandler

    def __init__(self, parent):
        if 'SimpleReflectivity' in parent.plugin_control.plugin_handler.get_loaded_plugins():
            parent.plugin_control.UnLoadPlugin_by_Name('SimpleReflectivity')
        framework.Template.__init__(self, parent)
        wx.EvtHandler.__init__(self)
        # self.parent = parent
        self.model_obj = self.GetModel()
        sample_panel = self.NewInputFolder('Sample')
        sample_sizer = wx.BoxSizer(wx.HORIZONTAL)
        sample_panel.SetSizer(sample_sizer)
        debug("Reflectivity: build sample widget")
        self.sample_widget = SamplePanel(sample_panel, self)
        sample_sizer.Add(self.sample_widget, 1, wx.EXPAND)
        sample_panel.Layout()

        simulation_panel = self.NewInputFolder('Simulations')
        simulation_sizer = wx.BoxSizer(wx.HORIZONTAL)
        simulation_panel.SetSizer(simulation_sizer)
        debug("Reflectivity: build data parameter widget")
        self.simulation_widget = DataParameterPanel(simulation_panel, self)
        simulation_sizer.Add(self.simulation_widget, 1, wx.EXPAND)
        simulation_panel.Layout()

        self.sample_widget.SetUpdateCallback(self.UpdateScript)
        self.simulation_widget.SetUpdateScriptFunc(self.UpdateScript)

        # Create the SLD plot
        sld_plot_panel = self.NewPlotFolder('SLD')
        sld_sizer = wx.BoxSizer(wx.HORIZONTAL)
        sld_plot_panel.SetSizer(sld_sizer)
        debug("Reflectivity: build SLD plot")
        self.sld_plot = SamplePlotPanel(sld_plot_panel, self)
        sld_sizer.Add(self.sld_plot, 1, wx.EXPAND)
        sld_plot_panel.Layout()

        if self.model_obj.script!='':
            if self.model_obj.filename!='':
                iprint("Reflectivity plugin: Reading loaded model")
                try:
                    self.ReadModel()
                except Exception as e:
                    iprint("Reflectivity plugin model could not be read.")
                    self.Remove()
                    raise e
            else:
                try:
                    self.ReadModel()
                except:
                    iprint("Reflectivity plugin: Creating new model")
                    self.CreateNewModel()
        else:
            iprint("Reflectivity plugin: Creating new model")
            self.CreateNewModel()

        # Create a menu for handling the plugin
        menu = self.NewMenu('Reflec')
        self.mb_export_sld = wx.MenuItem(menu, wx.NewId(),
                                         "Export SLD...",
                                         "Export the SLD to a ASCII file",
                                         wx.ITEM_NORMAL)
        menu.Append(self.mb_export_sld)
        self.mb_show_imag_sld = wx.MenuItem(menu, wx.NewId(),
                                            "Show Im SLD",
                                            "Toggles showing the imaginary part of the SLD",
                                            wx.ITEM_CHECK)
        menu.Append(self.mb_show_imag_sld)
        self.mb_generate_uncertainty = wx.MenuItem(menu, wx.NewId(),
                                                   "Uncertainty Profile...",
                                                   "Generate a plot showing SLD uncertainty after a fit",
                                                   wx.ITEM_NORMAL)
        menu.Append(self.mb_generate_uncertainty)
        self.mb_export_uncertainty = wx.MenuItem(menu, wx.NewId(),
                                                 "Export Uncertainty...",
                                                 "Export SLD uncertainty after a fit",
                                                 wx.ITEM_NORMAL)
        menu.Append(self.mb_export_uncertainty)
        menu.AppendSeparator()
        self.mb_show_imag_sld.Check(self.sld_plot.opt.show_imag)
        self.mb_autoupdate_sld = wx.MenuItem(menu, wx.NewId(),
                                             "Autoupdate SLD",
                                             "Toggles autoupdating the SLD during fitting",
                                             wx.ITEM_CHECK)
        menu.Append(self.mb_autoupdate_sld)
        self.mb_autoupdate_sld.Check(False)
        # self.mb_autoupdate_sld.SetCheckable(True)
        self.parent.Bind(wx.EVT_MENU, self.OnExportSLD, self.mb_export_sld)
        self.parent.Bind(wx.EVT_MENU, self.OnGenerateUncertainty, self.mb_generate_uncertainty)
        self.parent.Bind(wx.EVT_MENU, self.OnExportUncertainty, self.mb_export_uncertainty)
        self.parent.Bind(wx.EVT_MENU, self.OnAutoUpdateSLD, self.mb_autoupdate_sld)
        self.parent.Bind(wx.EVT_MENU, self.OnShowImagSLD, self.mb_show_imag_sld)
        self.parent.model_control.Bind(EVT_UPDATE_SCRIPT, self.ReadUpdateModel)
        self.StatusMessage('Reflectivity plugin loaded')

    def SetModelScript(self, script):
        framework.Template.SetModelScript(self, script)
        self._last_script = script

    def UpdateScript(self, event):
        self.WriteModel()

    def OnAutoUpdateSLD(self, evt):
        # self.mb_autoupdate_sld.Check(not self.mb_autoupdate_sld.IsChecked())
        pass

    def OnShowImagSLD(self, evt):
        self.sld_plot.opt.show_imag = self.mb_show_imag_sld.IsChecked()
        self.sld_plot.WriteConfig()
        self.sld_plot.Plot()

    @property
    def show_imag_sld(self):
        return self.sld_plot.opt.show_imag

    def OnExportSLD(self, evt):
        dlg = wx.FileDialog(self.parent, message="Export SLD to ...",
                            defaultFile="",
                            wildcard="Dat File (*.dat)|*.dat",
                            style=wx.FD_SAVE | wx.FD_CHANGE_DIR
                            )
        if dlg.ShowModal()==wx.ID_OK:
            fname = dlg.GetPath()
            result = True
            if os.path.exists(fname):
                filepath, filename = os.path.split(fname)
                result = self.ShowQuestionDialog('The file %s already exists.'
                                                 ' Do'
                                                 ' you wish to overwrite it?'
                                                 %filename)
            if result:
                try:
                    self.sld_plot.SavePlotData(fname)
                except IOError as e:
                    self.ShowErrorDialog(e.__str__())
                except Exception as e:
                    outp = io.StringIO()
                    traceback.print_exc(200, outp)
                    val = outp.getvalue()
                    outp.close()
                    self.ShowErrorDialog('Could not save the file.'
                                         ' Python Error:\n%s'%(val,))
        dlg.Destroy()

    def OnGenerateUncertainty(self, evt):
        validators = {
            'Reference Surface': ValueValidator(int),
            'Number of Samples': ValueValidator(int)
            }
        vals = {
            'Reference Surface': 0,
            'Number of Samples': 1000
            }
        pars = ['Reference Surface', 'Number of Samples']
        dlg = ValidateDialog(self.parent, pars, vals, validators,
                             title='Parameters for uncertainty graph')

        # Show the dialog
        if dlg.ShowModal()==wx.ID_OK:
            vals = dlg.GetValues()
            dlg.Destroy()
        else:
            dlg.Destroy()
            return

        model: Model = self.GetModel()
        model.simulate()
        self.sld_plot.PlotConfidence(int(vals['Reference Surface']), int(vals['Number of Samples']))

    def OnExportUncertainty(self, evt):
        dlg = wx.FileDialog(self.parent, message="Export SLD to ...",
                            defaultFile="",
                            wildcard="Dat File (*.dat)|*.dat",
                            style=wx.FD_SAVE | wx.FD_CHANGE_DIR
                            )
        if dlg.ShowModal()!=wx.ID_OK:
            dlg.Destroy()
            return

        fname = dlg.GetPath()
        dlg.Destroy()

        result = True
        if os.path.exists(fname):
            filepath, filename = os.path.split(fname)
            result = self.ShowQuestionDialog('The file %s already exists.'
                                             ' Do'
                                             ' you wish to overwrite it?'
                                             %filename)
        if not result:
            return

        validators = {
            'Reference Surface': ValueValidator(int),
            'Number of Samples': ValueValidator(int)
            }
        vals = {
            'Reference Surface': 0,
            'Number of Samples': 1000
            }
        pars = ['Reference Surface', 'Number of Samples']
        dlg = ValidateDialog(self.parent, pars, vals, validators,
                             title='Parameters for uncertainty graph')

        # Show the dialog
        if dlg.ShowModal()==wx.ID_OK:
            vals = dlg.GetValues()
            dlg.Destroy()
        else:
            dlg.Destroy()
            return

        model: Model = self.GetModel()
        model.simulate()
        self.sld_plot.SaveConfidenceData(fname, int(vals['Reference Surface']), int(vals['Number of Samples']),
                                         do_plot=True)

    def OnNewModel(self, event):
        ''' Create a new model
        '''
        dlg = wx.SingleChoiceDialog(self.parent, 'Choose a model type to use',
                                    'Models', avail_models,
                                    wx.CHOICEDLG_STYLE
                                    )

        if dlg.ShowModal()==wx.ID_OK:
            self.CreateNewModel('models.%s'%dlg.GetStringSelection())
        dlg.Destroy()

    def OnDataChanged(self, event):
        ''' Take into account changes in data..
        '''
        if event.new_model:
            return

        if event.data_moved or event.deleted or event.new_data or event.name_change:
            names = [data_set.name for data_set in self.GetModel().get_data()]
            self.simulation_widget.SetDataList(names)

            expl = self.simulation_widget.GetExpressionList()

            if len(names)-len(expl)==1:
                # Data set has been added:
                expl.append([])
                self.insert_new_data_segment(len(expl)-1)

            sims, insts, args = self.simulation_widget.GetSimArgs()

            if event.deleted:
                pos = list(range(len(expl)))
                [self.remove_data_segment(pos[-index-1]) for index in \
                 range(len(event.position))]
                [expl.pop(index) for index in event.position]
                [sims.pop(index) for index in event.position]
                [insts.pop(index) for index in event.position]
                [args.pop(index) for index in event.position]
            elif event.data_moved:
                if event.up:
                    # Moving up
                    for pos in event.position:
                        tmp = expl.pop(pos)
                        expl.insert(pos-1, tmp)
                        tmp = sims.pop(pos)
                        sims.insert(pos-1, tmp)
                        tmp = insts.pop(pos)
                        insts.insert(pos-1, tmp)
                        tmp = args.pop(pos)
                        args.insert(pos-1, tmp)
                else:
                    # Moving down...
                    for pos in event.position:
                        tmp = expl.pop(pos)
                        expl.insert(pos+1, tmp)
                        tmp = sims.pop(pos)
                        sims.insert(pos+1, tmp)
                        tmp = insts.pop(pos)
                        insts.insert(pos+1, tmp)
                        tmp = args.pop(pos)
                        args.insert(pos+1, tmp)

            self.simulation_widget.SetSimArgs(sims, insts, args)
            self.simulation_widget.SetExpressionList(expl)

            # Check so we have not clicked on new model button
            if self.GetModel().script!='':
                self.WriteModel()
                self.simulation_widget.UpdateListbox()
                if event.name_change:
                    self.sld_plot.Plot()
            else:
                self.simulation_widget.UpdateListbox(update_script=True)
        else:
            if event.data_changed:
                self.sld_plot.Plot()

    def OnOpenModel(self, event):
        '''
        Loads the sample into the plugin...
        '''
        self.mb_show_imag_sld.Check(self.sld_plot.opt.show_imag)
        self.ReadModel()

    def OnSimulate(self, event):
        '''
        Updates stuff after simulation
        '''
        if not self.mb_autoupdate_sld.IsChecked():
            # Calculate and update the sld plot, don't repeat when OnFittingUpdate does the call already
            wx.CallAfter(self.sld_plot.Plot)

    def OnFittingUpdate(self, event):
        '''
        Updates stuff during fitting
        '''
        # Calculate and update the sld plot
        if self.mb_autoupdate_sld.IsChecked():
            wx.CallAfter(self.sld_plot.Plot)

    def OnGridChange(self, event):
        """ Updates the simualtion panel when the grid changes

        :param event:
        :return:
        """
        self.sample_widget.Update(update_script=False)

    def InstrumentNameChange(self, old_name, new_name):
        '''OnInstrumentNameChange --> None
        
        Exchanges old_name to new name in the simulaitons.
        '''
        self.simulation_widget.InstrumentNameChange(old_name, new_name)

    def CreateNewModel(self, modelname='models.spec_nx'):
        '''Init the script in the model to yield the 
        correct script for initilization
        '''
        model_data = self.GetModel().get_data()
        nb_data_sets = len(model_data)

        script = self.GetNewModelScript(modelname, nb_data_sets)
        self.BuildNewModel(script)

        self.sample_widget.set_sampleh(self.sampleh)
        self.sample_widget.set_model(self.model)
        instrument = self.model.Instrument()
        self.sample_widget.SetInstrument({'inst': instrument})

        names = [data_set.name for data_set in model_data]
        self.simulation_widget.SetDataList(names)
        self.simulation_widget.SetParameterList([])
        # An empty list to the expression widget...
        self.simulation_widget.SetExpressionList([[] for item in names])
        self.simulation_widget.SetSimArgs(['Specular']*nb_data_sets,
                                          ['inst']*nb_data_sets,
                                          [['d.x'] for i in range(nb_data_sets)])
        self.simulation_widget.UpdateListbox(update_script=True)

        self.sample_widget.Update(update_script=True)
        # self.WriteModel()

    def WriteModel(self):
        parameter_list = self.simulation_widget.GetParameterList()
        sim_funcs, sim_insts, sim_args = self.simulation_widget.GetSimArgs()
        expression_list = self.simulation_widget.GetExpressionList()
        instruments = self.sample_widget.instruments

        self.write_model_script(sim_funcs, sim_insts, sim_args,
                                expression_list, parameter_list, instruments)
        try:
            self.SetXAxis(instruments[sim_insts[0]])
        except AttributeError:
            pass

    def SetXAxis(self, instrument):
        if self.previous_xaxis==instrument.coords:
            return
        coords = instrument.coords
        from genx import data
        if coords=='q':
            data.DataSet.simulation_params[0] = 0.001
            data.DataSet.simulation_params[1] = 0.601
        else:
            data.DataSet.simulation_params[0] = 0.01
            data.DataSet.simulation_params[1] = 6.01
        for ds in self.parent.model.data:
            ds.run_command()

    def AppendSim(self, sim_func, inst, args):
        self.simulation_widget.AppendSim(sim_func, inst, args)

    def ReadUpdateModel(self, evt):
        try:
            self.ReadModel(verbose=False)
        except GenxError:
            pass
        except Exception as e:
            self.StatusMessage(f'could not analyze script: {e}')

    def ReadModel(self, reevaluate=False, verbose=True):
        '''
        Reads in the current model and locates layers and stacks
        and sample defined inside BEGIN Sample section.
        '''
        if verbose: self.StatusMessage('Compiling the script...')
        self.CompileScript()
        if verbose: self.StatusMessage('Script compiled!')

        if verbose: self.StatusMessage('Trying to interpret the script...')

        instrument_names = self.find_instrument_names()

        if len(instrument_names)==0:
            self.ShowErrorDialog('Could not find any Instruments in the'+ \
                                 ' model script. Check the script.')
            self.StatusMessage('ERROR No Instruments in script')
            return

        if not 'inst' in instrument_names:
            self.ShowErrorDialog('Could not find the default'+
                                 ' Instrument, inst, in the'+
                                 ' model script. Check the script.')
            self.StatusMessage('ERROR No Instrument called inst in script')
            return

        sample_text = self.find_sample_section()

        if sample_text is None:
            self.ShowErrorDialog('Could not find the sample section'+ \
                                 ' in the model script.\n Can not load the sample in the editor.')
            self.StatusMessage('ERROR No sample section in script')
            return

        all_names, layers, stacks = self.find_layers_stacks(sample_text)

        if len(layers)==0:
            self.ShowErrorDialog('Could not find any Layers in the'+ \
                                 ' model script. Check the script.')
            self.StatusMessage('ERROR No Layers in script')
            return

        # Now its time to set all the parameters so that we have the strings
        # instead of the evaluated value - looks better
        for lay in layers:
            for par in lay[1].split(','):
                vars = par.split('=')
                exec('%s.%s = "%s"'%(lay[0], vars[0].strip(), vars[1].strip()), self.GetModel().script_module.__dict__)
        try:
            data_names, insts, sim_args, sim_exp, sim_funcs = self.find_sim_function_parameters()
        except LookupError:
            self.ShowErrorDialog('Could not locate all data sets in the'
                                 ' script. There should be %i datasets'%len(self.GetModel().get_data()))
            self.StatusMessage('ERROR No Layers in script')
            return

        uservars_lines = self.find_user_parameters()

        self.model = self.GetModel().script_module.model
        sample = self.GetModel().script_module.sample

        self.sampleh = SampleHandler(sample, all_names)
        self.sampleh.set_model(self.model)
        self.sample_widget.set_sampleh(self.sampleh)
        self.sample_widget.set_model(self.model)
        instruments = {}
        for name in instrument_names:
            instruments[name] = getattr(self.GetModel().script_module, name)
        self.sample_widget.SetInstrument(instruments)

        self.simulation_widget.SetDataList(data_names)
        self.simulation_widget.SetExpressionList(sim_exp)
        self.simulation_widget.SetParameterList(uservars_lines)

        self.simulation_widget.SetSimArgs(sim_funcs, insts, sim_args)

        self.sample_widget.Update(update_script=False)
        self.simulation_widget.UpdateListbox(update_script=False)
        # The code have a tendency to screw up the model slightly when compiling it - the sample will be connected
        # to the module therefore reset the compiled flag so that the model has to be recompiled before fitting.
        self.GetModel().compiled = False
        if reevaluate:
            if verbose: self.StatusMessage('Model analyzed and plugin updated!')
        else:
            if verbose: self.StatusMessage('New sample loaded to plugin!')
        self._last_script = self.model_obj.script

        # Setup the plot x-axis and simulation standard
        try:
            self.SetXAxis(self.sample_widget.instruments[instrument_names[0]])
        except AttributeError:
            pass



"""
Implementation of the iterative algorithm to define bi-layer periods of super-mirrors
that was introduced by J.B. Hayter and H.A. Mook in J. Appl. Cryst. (1989). 22, 35-41.

Adopted from Visual Basic script of PSI optics group.
"""

from numpy import *

SCALE_CONSTANT=3./sqrt(8.)

def delta_tau(tau, delta_sign, rho_fraction, zeta):
    # translated from VB script
    omega1 = sqrt(tau**2 - 1)
    omega2 = sqrt(tau**2 - rho_fraction)

    contrast = (omega2 - omega1) / (omega1 + omega2)
    kappa = (1 - contrast) / (1 + contrast)

    nu = log(1 - zeta) / (2*log(kappa))

    if nu<1: raise ValueError("Nu should be below 1, check input parameters. "
                              "This can be cause by a zeta value that is too smale, try e.g. 0.95.")

    rho_bar = (1 - abs(kappa)**(1/nu)) / (1 + abs(kappa)**(1/nu))
    arg = rho_bar * SCALE_CONSTANT

    d_omega1 = 2 * omega1 * arctan(arg / sqrt(1 - arg**2)) / pi
    wurzel = sqrt(1 + (omega1 + (delta_sign * d_omega1)) * (omega1 + (delta_sign * d_omega1)))

    return delta_sign * (wurzel - tau)

def sm_layers(rho1, rho2, N, zeta):
    tau = 1.1
    wert = SCALE_CONSTANT

    rho_fraction = rho2/rho1 # bruch

    D_SL=[]

    for i in range(N):
        # execute Neutrons method
        epsilon=1.
        while epsilon>1e-4:
            temp1 = tau - delta_tau(tau, -1, rho_fraction, zeta) - wert
            temp2 = 1e4 *  (1.0001 * tau - delta_tau((1.0001 * tau), -1, rho_fraction, zeta) - wert - temp1) / tau
            temp3 = temp1 / temp2
            tau = tau - temp3
            epsilon = abs(temp3/tau)

        wert = tau + delta_tau(tau, 1, rho_fraction, zeta)

        omega1 = sqrt(tau**2 - 1)
        omega2 = sqrt(tau**2 - rho_fraction)

        d_1 = sqrt(pi) / (4 * omega1 * sqrt(rho1))
        d_2 = sqrt(pi) / (4 * omega2 * sqrt(rho1))

        D_SL.append([d_1, d_2])

    return D_SL



"""
The model controller and socket server for remote optimization.
"""

import asyncio
import struct
from logging import debug, info, warning
from hashlib import blake2b

from . import messaging
from ..core import AUTH_SIZE, HANDSHAKE1, HANDSHAKE2
from ..model_control import ModelController
from ..diffev import DiffEv
from ..solver_basis import GenxOptimizerCallback, SolverParameterInfo, SolverResultInfo, SolverUpdateInfo


try:
    from mpi4py import MPI
except ImportError:
    rank = 0
    __mpi__ = False
else:
    __mpi__ = bool(MPI.COMM_WORLD.Get_size()>1)
    rank = MPI.COMM_WORLD.Get_rank()
    comm = MPI.COMM_WORLD


class RemotCallback(GenxOptimizerCallback):
    loop: asyncio.AbstractEventLoop = None

    def __init__(self, parent):
        GenxOptimizerCallback.__init__(self)
        self.parent = parent

    async def send_message(self, message: messaging.GenXMessage):
        self.parent.writer.write(message.message())
        await self.parent.writer.drain()

    def _send_message(self, message: messaging.GenXMessage):
        if self.loop is None:
            raise RuntimeError('Could not send message, no asyncio event loop defined')
        else:
            debug(f'sending message {message}')
            asyncio.run_coroutine_threadsafe(self.send_message(message), self.loop)

    def _end_fit(self):
        if self.loop is None:
            raise RuntimeError('Could not send message, no asyncio event loop defined')
        else:
            debug(f'cleanup fit locally')
            asyncio.run_coroutine_threadsafe(self.parent.cleanup(), self.loop)

    def text_output(self, text):
        msg = messaging.StingMessage(text)
        self._send_message(msg)

    def plot_output(self, update_data: SolverUpdateInfo):
        msg = messaging.OptimizerUpdate(update_data)
        self._send_message(msg)

    def parameter_output(self, param_info: SolverParameterInfo):
        msg = messaging.OptimizerUpdate(param_info)
        self._send_message(msg)

    def fitting_ended(self, result_data: SolverResultInfo):
        msg = messaging.OptimizerUpdate(result_data)
        self._send_message(msg)
        self._end_fit()

    def autosave(self):
        pass


class RemoteController(ModelController):
    """
    A model controller to execute on a server. Use RemoteController.serve with asyncio.
    """
    reader = None
    writer = None
    cleanup_phase = False

    key = b'empty'

    def __init__(self):
        ModelController.__init__(self, DiffEv())
        self.lock = None
        if rank==0:
            self.callbacks = RemotCallback(self)
            self.set_callbacks(self.callbacks)

    async def serve(self, address, port):
        self.lock = asyncio.locks.Lock()

        if rank==0:
            server = await asyncio.start_server(
                self.handle_connection, address, port)

            async with server:
                info(f"Starting listening on {address} with port={port}")
                await server.serve_forever()
        else:
            debug(f'Starting MPI client process for server on {address} with port={port}')
            await self.recive_mpi()

    async def handle_connection(self, reader, writer):
        debug('New connection')
        async with self.lock:
            debug('Connection setup and handshake')
            self.reader = reader
            self.writer = writer
            await self.handshake()
            self.cleanup_phase = False
            while self.reader is not None and not self.cleanup_phase:
                await self.recv_messages()

    async def handshake(self):
        key = self.key
        ref1 = blake2b(HANDSHAKE1, key=key, digest_size=AUTH_SIZE).hexdigest().encode('ascii')
        ref2 = blake2b(HANDSHAKE2, key=key, digest_size=AUTH_SIZE).hexdigest().encode('ascii')
        res = await self.reader.read(len(ref1))
        if res==ref1:
            debug('Incoming message correct, sending response.')
            self.writer.write(ref2)
            await self.writer.drain()
        else:
            debug("Handshake failed")
            self.writer.close()
            await self.writer.wait_closed()
            debug("Connection closed")
            self.reader = None
            self.writer = None

    async def recv_messages(self):
        try:
            res = await messaging.GenXMessage.receive(self.reader)
        except (ConnectionResetError, struct.error):
            if self.cleanup_phase:
                return
            else:
                warning("Connection was reset")
                await self.cleanup()
                return
        if isinstance(res, messaging.StingMessage):
            info(f"Received text message: {res.text}")
        elif isinstance(res, messaging.ActionMessage):
            if res.action_type is messaging.ActionType.START_FIT:
                info('Start fit was triggered')
                self.callbacks.loop = asyncio.get_running_loop()
                if self.optimizer.opt.use_mpi:
                    if __mpi__:
                        # server is just run on rank=0 so send model to clients
                        self.optimizer.opt.use_parallel_processing = False
                        self.optimizer.WriteConfig()
                        comm.Barrier()
                        comm.bcast(self.model, root=0)
                        comm.bcast(self.optimizer.opt, root=0)
                    else:
                        warning(
                            'Fit was started with use_mpi option selected, but server is started without MPI support')
                        self.optimizer.opt.use_mpi = False
                        self.optimizer.WriteConfig()
                self.StartFit()
            elif res.action_type is messaging.ActionType.STOP_FIT:
                info('Stop fit was triggered')
                await self.cleanup()
                self.callbacks.loop = None
            else:
                warning(f'Action not implemented {res!r}')
                return
        elif isinstance(res, messaging.ModelTransfer):
            info('Setting a new model')
            self.model = res.model
            self.optimizer.opt = res.fitparams
            self.optimizer.WriteConfig()
        elif isinstance(res, messaging.EchoMessage):
            info(f'Echoing message "{res.text}"')
            msg = messaging.StingMessage(res.text)
            await self.send_message(msg)

    async def send_message(self, message: messaging.GenXMessage):
        debug(f'Sending message {message}')
        self.writer.write(message.message())
        await self.writer.drain()

    async def recive_mpi(self):
        while True:
            comm.Barrier()
            self.model = comm.bcast(self.model, root=0)
            self.optimizer.opt = comm.bcast(self.optimizer.opt, root=0)
            self.optimizer.WriteConfig()
            self.StartFit()
            while self.optimizer.is_running():
                await asyncio.sleep(0.1)

    async def cleanup(self):
        if self.cleanup_phase:
            return
        debug("Starting cleanup sequence")
        self.cleanup_phase = True
        if self.optimizer.is_running():
            self.StopFit()
        while self.optimizer.is_running():
            await asyncio.sleep(0.1)
        debug("Closing connection")
        self.writer.close()
        await self.writer.wait_closed()
        debug("Connection closed")
        self.reader = None
        self.writer = None



'''
======================================================
:mod:`rigaku`  Rigaku ASCII file from Smartlab (*.ras)
======================================================

Loads the data exported by Rigaku Smartlab.
'''

import numpy as np

from ..data_loader_framework import Template
from ..utils import ShowWarningDialog

try:
    import wx
    from wx.lib.masked import NumCtrl
except ImportError:
    class void():
        pass

    wx=void()
    wx.Dialog=void

class Plugin(Template):
    wildcard='*.ras'

    def __init__(self, parent):
        Template.__init__(self, parent)
        self.x_col=0
        self.y_col=1
        self.e_col=1
        self.xe_col=-1
        self.comment='*'
        self.skip_rows=0
        self.delimiter=None

    def ReadHeader(self, fh):
        header = {}
        hl = fh.readline()
        while hl.startswith('*'):
            if hl.startswith('*RAS_INT_START'):
                break
            if '"' in hl:
                keys, value = hl.split('"', 1)
                value = value.split('"')[0]
            else:
                keys = hl; value = None
            hl = fh.readline()

            if value and '|' in value:
                value = value.split("|", 1)[-1]
            try:
                value = float(value)
            except (ValueError, TypeError):
                pass
            keys = keys[1:].lower().strip().split('_')
            if keys[0] not in header:
                header[keys[0]]={}
            subhdr = header[keys.pop(0)]
            while len(keys)>1:
                if keys[0] not in subhdr:
                    subhdr[keys[0]]={}
                elif not isinstance(subhdr[keys[0]], dict):
                    subhdr[keys[0]]={'_value': subhdr[keys[0]]}
                subhdr = subhdr[keys.pop(0)]
            subhdr[keys.pop(0)] = value
        del(header['ras'])
        return header

    def get_from_header(self, header, keys):
        value = header.get(keys[0], None)
        if value is None or len(keys)==1:
            return value
        else:
            if isinstance(value, dict):
                return self.get_from_header(value, keys[1:])
            else:
                return None

    def LoadData(self, dataset, filename, data_id=0):
        '''
        Loads the data from filename into the data_item_number.
        '''
        try:
            with open(filename, encoding='utf-8', errors='ignore') as fh:
                header = self.ReadHeader(fh)
                load_array=np.loadtxt(fh, delimiter=self.delimiter, comments=self.comment, skiprows=self.skip_rows)
        except Exception as e:
            ShowWarningDialog(self.parent, 'Could not load the file: '+ \
                              filename+' \nPlease check the format.\n\n numpy.loadtxt' \
                              +' gave the following error:\n'+str(e))
        else:
            dataset.x_raw=load_array[:, self.x_col]
            dataset.y_raw=load_array[:, self.y_col]
            dataset.error_raw=np.sqrt(load_array[:, self.y_col])
            if self.get_from_header(header, ['meas', 'scan', 'axis', 'x', '_value']) == "Omega/2-Theta":
                dataset.x_command = '2*x'
            else:
                dataset.x_command = 'x'
            # Run the commands on the data - this also sets the x,y, error memebers
            # of that data item.
            dataset.run_command()

            # insert metadata into ORSO compatible fields
            dataset.meta['data_source']['experiment']['instrument']='RAS'
            dataset.meta['data_source']['experiment']['probe']='xray'
            dataset.meta['data_source']['measurement']['scheme']='angle-dispersive'
            inst = dataset.meta['data_source']['measurement']['instrument_settings']
            inst['incident_angle'] = {'min': float(dataset.x_raw.min()), 'max': float(dataset.x_raw.max()),
                                      'unit': self.get_from_header(header, ['meas', 'scan', 'unit', 'x'])}
            inst['wavelength'] = {'magnitude': self.get_from_header(header, ['hw', 'xg', 'wave', 'length', 'alpha1']),
                                  'unit': 'angstrom'}
            dataset.meta['data_source']['sample'] = {'name': self.get_from_header(header, ['file', 'sample'])}
            # insert all metadata into user field
            dataset.meta['ras_header'] = header



"""
GUI support classes for the Reflectivity plug-in.
"""
import wx.html

from genx.core.custom_logging import iprint
from genx.model import Model
from genx.plugins.utils import ShowQuestionDialog, ShowWarningDialog
from . import reflectivity_images as images
from .custom_dialog import ComplexObjectValidator, FloatObjectValidator, NoMatchValidTextObjectValidator, \
    ParameterExpressionCombo, TextObjectValidator, ValidateDialog, ValidateFitDialog, \
    ValidateFitNotebookDialog
from .reflectivity_misc import ReflectivityModule
from .reflectivity_utils import find_code_segment, SampleHandler

_set_func_prefix='set'

class MyHtmlListBox(wx.html.HtmlListBox):

    def __init__(self, parent, id, size=(-1, -1), style=wx.BORDER_SUNKEN):
        wx.html.HtmlListBox.__init__(self, parent, id, size=size,
                                     style=style)
        self.SetItemList(['Starting up...'])

    def SetItemList(self, list):
        self.html_items = list
        self.SetItemCount(len(list))
        # self.RefreshAll()
        self.Refresh()

    def OnGetItem(self, n):
        return self.html_items[n]


class SamplePanel(wx.Panel):
    sampleh: SampleHandler
    model: ReflectivityModule

    def __init__(self, parent, plugin, refindexlist=None):
        wx.Panel.__init__(self, parent)
        if refindexlist is None:
            refindexlist = []
        self.refindexlist = refindexlist
        self.plugin = plugin
        self.variable_span = 0.25

        # Colours indicating different states
        # Green wx.Colour(138, 226, 52), ORANGE wx.Colour(245, 121, 0)
        self.fit_colour = (245, 121, 0)
        # Tango Sky blue wx.Colour(52, 101, 164), wx.Colour(114, 159, 207)
        self.const_fit_colour = (114, 159, 207)

        boxver = wx.BoxSizer(wx.HORIZONTAL)
        boxhor = wx.BoxSizer(wx.VERTICAL)
        self.toolbar = wx.ToolBar(self, style=wx.TB_FLAT | wx.TB_HORIZONTAL)
        self.do_toolbar()
        boxhor.Add(self.toolbar, proportion=0, flag=wx.EXPAND)
        boxhor.AddSpacer(2)
        self.listbox = MyHtmlListBox(self, -1, style=wx.BORDER_SUNKEN)
        # self.listbox.SetItemList(self.sampleh.getStringList())
        self.Bind(wx.EVT_LISTBOX_DCLICK, self.lbDoubleClick, self.listbox)
        boxhor.Add(self.listbox, 1, wx.EXPAND)

        boxver.Add(boxhor, 1, wx.EXPAND)

        self.SetSizer(boxver)
        self.toolbar.Realize()
        self.update_callback = lambda event: ''

    def do_toolbar(self):
        dpi_scale_factor = wx.GetApp().dpi_scale_factor
        tb_bmp_size = int(dpi_scale_factor*20)

        newid = wx.NewId()
        self.toolbar.AddTool(newid, 'Insert Layer',
                             bitmap=wx.Bitmap(images.insert_layer.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                             shortHelp='Insert a Layer')
        self.Bind(wx.EVT_TOOL, self.InsertLay, id=newid)

        newid = wx.NewId()
        self.toolbar.AddTool(newid, 'Insert Stack',
                             bitmap=wx.Bitmap(images.insert_stack.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                             shortHelp='Insert a Stack')
        self.Bind(wx.EVT_TOOL, self.InsertStack, id=newid)

        newid = wx.NewId()
        self.toolbar.AddTool(newid, 'Delete',
                             bitmap=wx.Bitmap(images.delete.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                             shortHelp='Delete item')
        self.Bind(wx.EVT_TOOL, self.DeleteSample, id=newid)

        newid = wx.NewId()
        self.toolbar.AddTool(newid, 'Rename',
                             bitmap=wx.Bitmap(images.change_name.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                             shortHelp='Rename')
        self.Bind(wx.EVT_TOOL, self.ChangeName, id=newid)

        newid = wx.NewId()
        self.toolbar.AddTool(newid, 'Move up',
                             bitmap=wx.Bitmap(images.move_up.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                             shortHelp='Move item up')
        self.Bind(wx.EVT_TOOL, self.MoveUp, id=newid)

        newid = wx.NewId()
        self.toolbar.AddTool(newid, 'Move down',
                             bitmap=wx.Bitmap(images.move_down.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                             shortHelp='Move item down')
        self.Bind(wx.EVT_TOOL, self.MoveDown, id=newid)

        newid = wx.NewId()
        self.toolbar.AddTool(newid, 'Edit Sample',
                             bitmap=wx.Bitmap(images.sample.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                             shortHelp='Edit Sample parameters')
        self.Bind(wx.EVT_TOOL, self.EditSampleParameters, id=newid)

        newid = wx.NewId()
        self.toolbar.AddTool(newid, 'Edit Instrument',
                             bitmap=wx.Bitmap(images.instrument.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                             shortHelp='Edit Instruments')
        self.Bind(wx.EVT_TOOL, self.EditInstrument, id=newid)

    def SetUpdateCallback(self, func):
        ''' SetUpdateCallback(self, func) --> None

        Sets the update callback will be called when the sample is updated.
        The call is on the form func(event)
        '''
        self.update_callback = func

    def set_sampleh(self, sampleh: SampleHandler):
        self.sampleh = sampleh

    def set_model(self, model: ReflectivityModule):
        self.model = model

    def create_html_decorator(self):
        """
        creates a html decorator function
        :return:
        """
        grid_parameters = self.plugin.GetModel().get_parameters()
        dic_lookup = {}
        for par in grid_parameters.get_names():
            l = par.split('.')
            if len(l)==2:
                name = l[0]
                par_name = l[1][3:].lower()
                dic_lookup[(name, par_name)] = (grid_parameters.get_value_by_name(par),
                                                grid_parameters.get_fit_state_by_name(par)
                                                )
        fit_color_str = "rgb(%d,%d,%d)"%self.fit_colour
        const_fit_color_str = "rgb(%d,%d,%d)"%self.const_fit_colour

        def decorator(name, str):
            """ Decorator to indicate the parameters that are fitted"""
            try:
                start_index = str.index('(')+1
            except ValueError:
                start_index = 0
            ret_str = str[:start_index]
            for par_str in str[start_index:].split(','):
                par_name = par_str.split('=')[0].strip()
                # par_name normal paramter (real number)
                if (name, par_name) in dic_lookup:
                    val, state = dic_lookup[(name, par_name)]
                    if state==1:
                        par_str = ' <font color=%s><b>%s=%.2e</b></font>,'%(fit_color_str, par_name, val)
                    elif state==2:
                        par_str = ' <font color=%s><b>%s=%.2e</b></font>,'%(const_fit_color_str, par_name, val)
                # par_name is a complex parameter...
                elif (name, par_name+'real') in dic_lookup or (name, par_name+'imag') in dic_lookup:
                    if (name, par_name+'real') in dic_lookup:
                        val, state = dic_lookup[(name, par_name+'real')]
                        if state==1:
                            par_str = ' <font color=%s><b>%s=(%.2e,</b></font>'%(fit_color_str, par_name, val)
                        elif state==2:
                            par_str = ' <font color=%s><b>%s=(%.2e,</b></font>'%(const_fit_color_str, par_name, val)
                    else:
                        par_str = ' <b>%s=??+</b>'%par_name
                    if (name, par_name+'imag') in dic_lookup:
                        val, state = dic_lookup[(name, par_name+'imag')]
                        if state==1:
                            par_str += ' <font color=%s><b>%.2e)</b></font>,'%(fit_color_str, val)
                        elif state==2:
                            par_str += ' <font color=%s><b>%.2e)</b></font>,'%(const_fit_color_str, val)
                    else:
                        par_str += ' <b>??)</b>,'

                else:
                    par_str += ','
                ret_str += par_str
            # Remove trailing ,
            if ret_str[-1]==',':
                ret_str = ret_str[:-1]
            if str[-1]==')' and ret_str[-1]!=')':
                ret_str += ')'
            return ret_str

        return decorator

    def Update(self, update_script=True):
        deco = self.create_html_decorator()
        sl = self.sampleh.getStringList(html_encoding=True, html_decorator=deco)
        self.listbox.SetItemList(sl)
        if update_script:
            self.update_callback(None)

    def SetSample(self, sample, names):
        self.sampleh.sample = sample
        self.sampleh.names = names
        self.Update()

    def EditSampleParameters(self, evt):
        """ Event handler that creates a dialog box to edit the sample parameters.

        :param evt:
        :return: Nothing
        """
        obj_name = 'sample'
        eval_func = self.plugin.GetModel().eval_in_model
        grid_parameters = self.plugin.GetModel().get_parameters()

        validators = {}
        vals = {}
        pars = []
        items = []
        editable = {}
        try:
            string_choices = self.model.sample_string_choices
        except Exception as e:
            string_choices = {}
        for item in self.model.SampleParameters:
            if item!='Stacks' and item!='Substrate' and item!='Ambient':
                if item in string_choices:
                    validators[item] = string_choices[item]
                else:
                    validators[item] = FloatObjectValidator()
                val = getattr(self.sampleh.sample, item)
                vals[item] = val
                pars.append(item)
                items.append((item, val))
                # Check if the parameter is in the grid and in that case set it as uneditable
                func_name = obj_name+'.'+_set_func_prefix+item.capitalize()
                grid_value = grid_parameters.get_value_by_name(func_name)
                editable[item] = grid_parameters.get_fit_state_by_name(func_name)
                if grid_value is not None:
                    vals[item] = grid_value
        try:
            groups = self.model.SampleGroups
        except Exception:
            groups = False
        try:
            units = self.model.SampleUnits
        except Exception:
            units = False

        dlg = ValidateFitDialog(self, pars, vals, validators,
                                title='Sample Editor', groups=groups,
                                units=units, editable_pars=editable)

        if dlg.ShowModal()==wx.ID_OK:
            old_vals = vals
            vals = dlg.GetValues()
            # print vals
            states = dlg.GetStates()
            for par in pars:
                if not states[par]:
                    old_type = type(old_vals[par])
                    setattr(self.sampleh.sample, par, old_type(vals[par]))
                if editable[par]!=states[par]:
                    value = eval_func(vals[par])
                    minval = min(value*(1-self.variable_span), value*(1+self.variable_span))
                    maxval = max(value*(1-self.variable_span), value*(1+self.variable_span))
                    func_name = obj_name+'.'+_set_func_prefix+par.capitalize()
                    grid_parameters.set_fit_state_by_name(func_name, value, states[par], minval, maxval)
                    # Tell the grid to reload the parameters
                    self.plugin.parent.paramter_grid.SetParameters(grid_parameters)

            self.Update()

        dlg.Destroy()

    def SetInstrument(self, instruments):
        '''SetInstrument(self, instrument) --> None

        Sets the instruments should be a dictionary of instruments with the key being the
        name of the instrument
        '''
        self.instruments = instruments

    def EditInstrument(self, evt):
        """Event handler that creates an dialog box to edit the instruments.

        :param evt:
        :return: Nothing
        """
        eval_func = self.plugin.GetModel().eval_in_model
        validators = {}
        vals = {}
        editable = {}
        grid_parameters = self.plugin.GetModel().get_parameters()
        for inst_name in self.instruments:
            vals[inst_name] = {}
            editable[inst_name] = {}

        pars = []
        model_inst_params = self.model.InstrumentParameters
        for item in model_inst_params:
            if item in self.model.instrument_string_choices:
                # validators.append(self.model.instrument_string_choices[item])
                validators[item] = self.model.instrument_string_choices[item]
            else:
                # validators.append(FloatObjectValidator())
                validators[item] = FloatObjectValidator()
            for inst_name in self.instruments:
                val = getattr(self.instruments[inst_name], item)
                vals[inst_name][item] = val
                # Check if the parameter is in the grid and in that case set it as uneditable
                func_name = inst_name+'.'+_set_func_prefix+item.capitalize()
                grid_value = grid_parameters.get_value_by_name(func_name)
                editable[inst_name][item] = grid_parameters.get_fit_state_by_name(func_name)
                if grid_value is not None:
                    vals[inst_name][item] = grid_value
            pars.append(item)

        old_insts = []
        for inst_name in self.instruments:
            old_insts.append(inst_name)

        try:
            groups = self.model.InstrumentGroups
        except Exception:
            groups = False
        try:
            units = self.model.InstrumentUnits
        except Exception:
            units = False
        dlg = ValidateFitNotebookDialog(self, pars, vals, validators,
                                        title='Instrument Editor', groups=groups,
                                        units=units, fixed_pages=['inst'], editable_pars=editable)

        if dlg.ShowModal()==wx.ID_OK:
            old_vals = vals
            vals = dlg.GetValues()
            # print vals
            states = dlg.GetStates()
            self.instruments = {}
            for inst_name in vals:
                new_instrument = False
                if inst_name not in self.instruments:
                    # A new instrument must be created:
                    self.instruments[inst_name] = self.model.Instrument()
                    new_instrument = True
                for par in self.model.InstrumentParameters:
                    if not states[inst_name][par]:
                        orig_type = type(model_inst_params[par])
                        if orig_type is str:
                            e_value = vals[inst_name][par]
                        else:
                            e_value = eval_func(vals[inst_name][par])
                        setattr(self.instruments[inst_name], par, orig_type(e_value))
                    else:
                        setattr(self.instruments[inst_name], par, old_vals[inst_name][par])
                    if new_instrument and states[inst_name][par]>0:
                        value = eval_func(vals[inst_name][par])
                        minval = min(value*(1-self.variable_span), value*(1+self.variable_span))
                        maxval = max(value*(1-self.variable_span), value*(1+self.variable_span))
                        func_name = inst_name+'.'+_set_func_prefix+par.capitalize()
                        grid_parameters.set_fit_state_by_name(func_name, value, states[inst_name][par], minval, maxval)
                    elif not new_instrument:
                        if editable[inst_name][par]!=states[inst_name][par]:
                            value = eval_func(vals[inst_name][par])
                            minval = min(value*(1-self.variable_span), value*(1+self.variable_span))
                            maxval = max(value*(1-self.variable_span), value*(1+self.variable_span))
                            func_name = inst_name+'.'+_set_func_prefix+par.capitalize()
                            grid_parameters.set_fit_state_by_name(func_name, value, states[inst_name][par], minval,
                                                                  maxval)

            # Loop to remove instrument from grid if not returned from Dialog
            for inst_name in old_insts:
                if inst_name not in list(vals.keys()):
                    for par in self.model.InstrumentParameters:
                        if editable[inst_name][par]>0:
                            func_name = inst_name+'.'+_set_func_prefix+par.capitalize()
                            grid_parameters.set_fit_state_by_name(func_name, 0, 0, 0, 0)

            # Tell the grid to reload the parameters
            self.plugin.parent.paramter_grid.SetParameters(grid_parameters)

            for change in dlg.GetChanges():
                if change[0]!='' and change[1]!='':
                    self.plugin.InstrumentNameChange(change[0], change[1])
                elif change[1]=='':
                    self.plugin.InstrumentNameChange(change[0], 'inst')

            self.Update()
        else:
            pass
        dlg.Destroy()

    def MoveUp(self, evt):
        sl = self.sampleh.moveUp(self.listbox.GetSelection())
        if sl:
            self.Update()
            self.listbox.SetSelection(self.listbox.GetSelection()-1)

    def MoveDown(self, evt):
        sl = self.sampleh.moveDown(self.listbox.GetSelection())
        if sl:
            self.Update()
            self.listbox.SetSelection(self.listbox.GetSelection()+1)

    def InsertStack(self, evt):
        # Create Dialog box
        validators = {'Name': NoMatchValidTextObjectValidator(self.sampleh.names)}
        vals = {}
        pars = ['Name']
        vals['Name'] = 'name'
        dlg = ValidateDialog(self, pars, vals, validators,
                             title='Give Stack Name')

        # Show the dialog
        if dlg.ShowModal()==wx.ID_OK:
            vals = dlg.GetValues()
        dlg.Destroy()
        # if not a value is selected operate on first
        pos = max(self.listbox.GetSelection(), 0)
        sl = self.sampleh.insertItem(pos, 'Stack', vals['Name'])
        if sl:
            self.Update()
        else:
            self.plugin.ShowWarningDialog('Can not insert a stack at the'
                                          ' current position.')

    def InsertLay(self, evt):
        # Create Dialog box
        # items = [('Name', 'name')]
        # validators = [NoMatchValidTextObjectValidator(self.sampleh.names)]
        dlg = ValidateDialog(self, ['Name'], {'Name': 'name'},
                             {'Name': NoMatchValidTextObjectValidator(self.sampleh.names)},
                             title='Give Layer Name')
        # Show the dialog
        if dlg.ShowModal()==wx.ID_OK:
            vals = dlg.GetValues()
        else:
            vals = {'Name': 'name'}
        dlg.Destroy()
        # if not a value is selected operate on first
        pos = max(self.listbox.GetSelection(), 0)
        # Create the Layer
        sl = self.sampleh.insertItem(pos, 'Layer', vals['Name'])
        if sl:
            self.Update()
        else:
            self.plugin.ShowWarningDialog('Can not insert a layer at the'
                                          ' current position. Layers has to be part of a stack.')

    def DeleteSample(self, evt):
        slold = self.sampleh.getStringList()
        sl = self.sampleh.deleteItem(self.listbox.GetSelection())
        if sl:
            self.Update()

    def ChangeName(self, evt):
        '''Change the name of the current selected item.
        '''
        pos = self.listbox.GetSelection()
        if pos==0 or pos==len(self.sampleh.names)-1:
            self.plugin.ShowInfoDialog('It is forbidden to change the'
                                       'name of the substrate (Sub) and the Ambient (Amb) layers.')
        else:
            unallowed_names = self.sampleh.names[:pos]+ \
                              self.sampleh.names[max(0, pos-1):]
            dlg = ValidateDialog(self, ['Name'], {'Name': self.sampleh.names[pos]},
                                 {'Name': NoMatchValidTextObjectValidator(unallowed_names)},
                                 title='Give New Name')

            if dlg.ShowModal()==wx.ID_OK:
                vals = dlg.GetValues()
                result = self.sampleh.changeName(pos, vals['Name'])
                if result:
                    self.Update()
                else:
                    iprint('Unexpected problems when changing name...')
            dlg.Destroy()

    def lbDoubleClick(self, evt):
        sel = self.sampleh.getItem(self.listbox.GetSelection())
        obj_name = self.sampleh.getName(self.listbox.GetSelection())
        eval_func = self.plugin.GetModel().eval_in_model
        sl = None
        items = []
        validators = {}
        vals = {}
        pars = []
        editable = {}
        grid_parameters = self.plugin.GetModel().get_parameters()
        if isinstance(sel, self.model.Layer):
            # The selected item is a Layer
            for item in list(self.model.LayerParameters.keys()):
                value = getattr(sel, item)
                vals[item] = value
                # if item!='n' and item!='fb':
                if type(self.model.LayerParameters[item])!=type(1+1.0J):
                    # Handle real parameters
                    validators[item] = FloatObjectValidator(eval_func, alt_types=[self.model.Layer])
                    func_name = obj_name+'.'+_set_func_prefix+item.capitalize()
                    grid_value = grid_parameters.get_value_by_name(func_name)
                    if grid_value is not None:
                        vals[item] = grid_value
                    editable[item] = grid_parameters.get_fit_state_by_name(func_name)

                else:
                    # Handle complex parameters
                    validators[item] = ComplexObjectValidator(eval_func, alt_types=[self.model.Layer])
                    func_name = obj_name+'.'+_set_func_prefix+item.capitalize()
                    grid_value_real = grid_parameters.get_value_by_name(func_name+'real')
                    grid_value_imag = grid_parameters.get_value_by_name(func_name+'imag')
                    if grid_value_real is not None:
                        v = eval_func(vals[item]) if type(vals[item]) is str else vals[item]
                        vals[item] = grid_value_real+v.imag*1.0J
                    if grid_value_imag is not None:
                        v = eval_func(vals[item]) if type(vals[item]) is str else vals[item]
                        vals[item] = v.real+grid_value_imag*1.0J
                    editable[item] = max(grid_parameters.get_fit_state_by_name(func_name+'real'),
                                         grid_parameters.get_fit_state_by_name(func_name+'imag'))

                items.append((item, value))
                pars.append(item)

                # Check if the parameter is in the grid and in that case set it as uneditable
                # func_name = obj_name + '.' + _set_func_prefix + item.capitalize()
                # grid_value = grid_parameters.get_value_by_name(func_name)
                # editable[item] = grid_parameters.get_fit_state_by_name(func_name)

            try:
                groups = self.model.LayerGroups
            except Exception:
                groups = False
            try:
                units = self.model.LayerUnits
            except Exception:
                units = False

            dlg = ValidateFitDialog(self, pars, vals, validators,
                                    title='Layer Editor', groups=groups,
                                    units=units, editable_pars=editable)

            if dlg.ShowModal()==wx.ID_OK:
                vals = dlg.GetValues()
                states = dlg.GetStates()
                for par in list(self.model.LayerParameters.keys()):
                    if not states[par]:
                        setattr(sel, par, vals[par])
                    if editable[par]!=states[par]:
                        value = eval_func(vals[par])

                        if type(self.model.LayerParameters[par]) is complex:
                            # print type(value)
                            func_name = obj_name+'.'+_set_func_prefix+par.capitalize()+'real'
                            val = value.real
                            minval = min(val*(1-self.variable_span), val*(1+self.variable_span))
                            maxval = max(val*(1-self.variable_span), val*(1+self.variable_span))
                            grid_parameters.set_fit_state_by_name(func_name, val, states[par], minval, maxval)
                            val = value.imag
                            minval = min(val*(1-self.variable_span), val*(1+self.variable_span))
                            maxval = max(val*(1-self.variable_span), val*(1+self.variable_span))
                            func_name = obj_name+'.'+_set_func_prefix+par.capitalize()+'imag'
                            grid_parameters.set_fit_state_by_name(func_name, val, states[par], minval, maxval)
                        else:
                            val = value
                            minval = min(val*(1-self.variable_span), val*(1+self.variable_span))
                            maxval = max(val*(1-self.variable_span), val*(1+self.variable_span))
                            func_name = obj_name+'.'+_set_func_prefix+par.capitalize()
                            grid_parameters.set_fit_state_by_name(func_name, value, states[par], minval, maxval)

                        # Does not seem to be necessary
                        self.plugin.parent.paramter_grid.SetParameters(grid_parameters)
                sl = self.sampleh.getStringList()
            dlg.Destroy()

        else:
            # The selected item is a Stack
            for item in list(self.model.StackParameters.keys()):
                if item!='Layers':
                    value = getattr(sel, item)
                    if isinstance(value, float):
                        validators[item] = FloatObjectValidator(eval_func, alt_types=[self.model.Stack])
                    else:
                        validators[item] = TextObjectValidator()
                    items.append((item, value))
                    pars.append(item)
                    vals[item] = value

                    # Check if the parameter is in the grid and in that case set it as uneditable
                    func_name = obj_name+'.'+_set_func_prefix+item.capitalize()
                    grid_value = grid_parameters.get_value_by_name(func_name)
                    editable[item] = grid_parameters.get_fit_state_by_name(func_name)
                    if grid_value is not None:
                        vals[item] = grid_value

            try:
                groups = self.model.StackGroups
            except Exception:
                groups = False
            try:
                units = self.model.StackUnits
            except Exception:
                units = False

            dlg = ValidateFitDialog(self, pars, vals, validators,
                                    title='Layer Editor', groups=groups,
                                    units=units, editable_pars=editable)
            if dlg.ShowModal()==wx.ID_OK:
                vals = dlg.GetValues()
                states = dlg.GetStates()
                for par in pars:
                    if not states[par]:
                        setattr(sel, par, vals[par])
                    if editable[par]!=states[par]:
                        value = eval_func(vals[par])
                        minval = min(value*(1-self.variable_span), value*(1+self.variable_span))
                        maxval = max(value*(1-self.variable_span), value*(1+self.variable_span))
                        func_name = obj_name+'.'+_set_func_prefix+par.capitalize()
                        grid_parameters.set_fit_state_by_name(func_name, value, states[par], minval, maxval)
                        # Does not seem to be necessary
                        self.plugin.parent.paramter_grid.SetParameters(grid_parameters)
                sl = self.sampleh.getStringList()

            dlg.Destroy()

        if sl:
            self.Update()


class DataParameterPanel(wx.Panel):
    ''' Widget that defines parameters coupling and different parameters
    for different data sets.
    '''

    def __init__(self, parent, plugin):
        wx.Panel.__init__(self, parent)
        self.plugin = plugin
        boxver = wx.BoxSizer(wx.VERTICAL)
        # Indention for a command - used to seperate commands and data
        self.command_indent = '<pre>   '
        self.script_update_func = None
        self.parameterlist = []

        self.toolbar = wx.ToolBar(self, style=wx.TB_FLAT | wx.TB_HORIZONTAL)
        self.do_toolbar()
        boxver.Add(self.toolbar, proportion=0, flag=wx.EXPAND, border=1)
        boxver.AddSpacer(4)

        self.listbox = MyHtmlListBox(self, -1, style=wx.BORDER_SUNKEN)
        self.Bind(wx.EVT_LISTBOX_DCLICK, self.Edit, self.listbox)
        boxver.Add(self.listbox, 1, wx.EXPAND)

        self.SetSizer(boxver)

    def do_toolbar(self):
        dpi_scale_factor = wx.GetApp().dpi_scale_factor
        tb_bmp_size = int(dpi_scale_factor*20)

        button_names = ['Insert', 'Delete', 'User Variables']
        button_images = [wx.Bitmap(images.add.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                         wx.Bitmap(images.delete.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                         wx.Bitmap(images.custom_parameters.GetImage().Scale(tb_bmp_size, tb_bmp_size))]
        callbacks = [self.Insert, self.Delete, self.EditPars]
        tooltips = ['Insert a command', 'Delete command', 'Edit user variables']

        for i in range(len(button_names)):
            newid = wx.NewId()
            self.toolbar.AddTool(newid, label=button_names[i],
                                 bitmap=button_images[i],
                                 shortHelp=tooltips[i])
            self.Bind(wx.EVT_TOOL, callbacks[i], id=newid)
        self.toolbar.Realize()

    def onsimulate(self, event):
        """ Function to simulate the model.

        :return:
        """
        self.plugin.parent.eh_tb_simulate(event)

    def SetDataList(self, datalist):
        '''SetDataList(self, datalist) --> None

        Sets the name of the available data sets
        '''
        self.datalist = datalist

    def GetDataList(self):
        '''SetDataList(self) --> list

        Retrives the data list
        '''
        return self.datalist

    def SetParameterList(self, parameterlist):
        '''SetParameterList(self, parameterlist) --> None

        Sets the code list for all definition of custom parameters
        '''
        self.parameterlist = parameterlist

    def GetParameterList(self):
        '''SetParameterList(self) --> list

        Retrives the parameter list
        '''
        return self.parameterlist

    def SetExpressionList(self, expressionlist):
        '''SetExpressionList(expressionlist) --> None

        Sets the expression list, should be a 2D list with the
        one list for each item in datalist
        '''
        if len(expressionlist)!=len(self.datalist):
            raise ValueError('The list of expression has to have the'+ \
                             ' same length as the data list')
        self.expressionlist = expressionlist

    def GetExpressionList(self):
        '''GetExpressionList(self) --> expression list

        Returns the expressionlist
        '''
        return self.expressionlist

    def SetSimArgs(self, sim_funcs, insts, args):
        '''SetSimArgs(self, sim_func, inst, args) --> None

        Sets the current simulation function for each data set
        their instruments and arguments
        sim_func: A list of simulation names
        inst: A list of instrument names
        args: A list of argument list for the sim_func's each argument should be a string.
        '''
        if len(sim_funcs)!=len(self.datalist):
            raise ValueError('The list of sim_funcs has to have the'+ \
                             ' same length as the data list')
        if len(insts)!=len(self.datalist):
            raise ValueError('The list of insts has to have the'+ \
                             ' same length as the data list')
        if len(args)!=len(self.datalist):
            raise ValueError('The list of args has to have the'+ \
                             ' same length as the data list')
        self.sim_funcs = sim_funcs[:]
        self.insts = insts[:]
        self.args = args[:]

    def GetSimArgs(self):
        '''GetSimArgs(self) --> (sim_funcs, insts, args)

        See SetSimArgs for a description of the parameters
        '''
        return self.sim_funcs, self.insts, self.args

    def AppendSim(self, sim_func, inst, args):
        '''AppendSim(self, sim_func, inst, args) --> None

        Appends a simultion to the Panel lists
        '''
        self.sim_funcs.append(sim_func)
        self.insts.append(inst)
        self.args.append(args)

    def InstrumentNameChange(self, old_name, new_name):
        '''OnInstrumentNameChange --> None

        Exchanges old_name to new name in the simulations.'''

        for i in range(len(self.insts)):
            if self.insts[i]==old_name:
                self.insts[i] = new_name
        self.update_listbox()

    def SetUpdateScriptFunc(self, func):
        '''SetUpdateScriptFunc(self, func) --> None

        Sets the function to be called when the script needs to be updated.
        will only be called as func(event)
        '''
        self.script_update_func = func

    def UpdateListbox(self, update_script=True):
        '''Update(self) --> None

        Update the listbox and runs the callback script_update_func
        '''
        self.update_listbox()
        if self.script_update_func and update_script:
            self.script_update_func(None)
        self.Refresh()

    def update_listbox(self):
        '''update_listbox(self) --> None

         updates the listbox.
        '''
        list_strings = []
        for i in range(len(self.datalist)):
            str_arg = ', '.join(self.args[i])
            list_strings.append('<code><b>%s</b>: %s(%s, %s)</code>'
                                ' \n'%(self.datalist[i],
                                       self.sim_funcs[i],
                                       str_arg, self.insts[i]))
            for item in self.expressionlist[i]:
                list_strings.append(self.command_indent+'%s</pre>'%item)

        self.listbox.SetItemList(list_strings)

    def get_expression_position(self):
        '''get_expression_position(self) --> (dataitem, expression)

        Finds the position in the expression list for a certain item.
        return -1 if it can not be found.
        '''
        index = self.listbox.GetSelection()

        if index==wx.NOT_FOUND:
            return -1, -1

        dataindex = -1
        itemindex = -1
        listindex = -1
        for i in range(len(self.datalist)):
            dataindex += 1
            listindex += 1
            itemindex = -1
            if listindex>=index:
                return dataindex, itemindex
            for item in self.expressionlist[i]:
                listindex += 1
                itemindex += 1
                if listindex>=index:
                    return dataindex, itemindex

        # If all other things fail...
        return -1, -1

    def Edit(self, event):
        '''Edit(self, event) --> None

        Edits an entry in the list
        '''
        data_pos, exp_pos = self.get_expression_position()
        if exp_pos!=-1 and data_pos!=-1:
            # Editing the expressions for variables
            list_item = self.expressionlist[data_pos][exp_pos]
            dlg = ParameterExpressionDialog(self, self.plugin.GetModel(),
                                            list_item, sim_func=self.onsimulate)
            if dlg.ShowModal()==wx.ID_OK:
                exp = dlg.GetExpression()
                self.expressionlist[data_pos][exp_pos] = exp
                self.UpdateListbox()
        if exp_pos==-1 and data_pos!=-1:
            # Editing the simulation function and its arguments
            dlg = SimulationExpressionDialog(self, self.plugin.GetModel(),
                                             self.plugin.sample_widget.instruments,
                                             self.sim_funcs[data_pos],
                                             self.args[data_pos],
                                             self.insts[data_pos], data_pos)
            if dlg.ShowModal()==wx.ID_OK:
                self.args[data_pos] = dlg.GetExpressions()
                self.insts[data_pos] = dlg.GetInstrument()
                self.sim_funcs[data_pos] = dlg.GetSim()
                self.UpdateListbox()

    def Insert(self, event):
        ''' Insert(self, event) --> None

        Inserts a new operations
        '''
        data_pos, exp_pos = self.get_expression_position()
        if data_pos!=-1:
            dlg = ParameterExpressionDialog(self, self.plugin.GetModel(), sim_func=self.onsimulate)
            if dlg.ShowModal()==wx.ID_OK:
                exp = dlg.GetExpression()
                if exp_pos==-1:
                    self.expressionlist[data_pos].insert(0, exp)
                else:
                    self.expressionlist[data_pos].insert(exp_pos, exp)
                self.UpdateListbox()

    def Delete(self, event):
        '''Delete(self, event) --> None

        Deletes an operation
        '''
        data_pos, exp_pos = self.get_expression_position()
        if exp_pos!=-1 and data_pos!=-1:
            self.expressionlist[data_pos].pop(exp_pos)
            self.UpdateListbox()

    def MoveUp(self, event):
        '''MoveUp(self, event) --> None

        Move an operation up
        '''
        pass

    def MoveDown(self, event):
        '''MoveDown(self, event) --> None

        Moves an operation down
        '''
        pass

    def EditPars(self, event):
        ''' EditPars(self, event) --> None

        Creates a new parameter
        '''
        dlg = EditCustomParameters(self, self.plugin.GetModel(),
                                   self.parameterlist)
        if dlg.ShowModal()==wx.ID_OK:
            self.parameterlist = dlg.GetLines()
            self.UpdateListbox()
        dlg.Destroy()

    def OnDataChanged(self, event):
        '''OnDataChanged(self, event) --> None

        Updated the data list
        '''
        self.UpdateListbox(update_script=False)


class EditCustomParameters(wx.Dialog):
    model: Model

    def __init__(self, parent, model, lines):
        wx.Dialog.__init__(self, parent, -1, 'Custom parameter editor')
        self.SetAutoLayout(True)
        self.model = model
        self.lines = lines
        self.var_name = 'cp'

        sizer = wx.BoxSizer(wx.VERTICAL)
        name_ctrl_sizer = wx.GridBagSizer(2, 3)

        col_labels = ['Name', 'Value', 'Sigma (for systematic error)']

        for item, index in zip(col_labels, list(range(len(col_labels)))):
            label = wx.StaticText(self, -1, item)
            name_ctrl_sizer.Add(label, (0, index), flag=wx.ALIGN_LEFT, border=5)

        self.name_ctrl = wx.TextCtrl(self, -1, size=(120, -1))
        name_ctrl_sizer.Add(self.name_ctrl, (1, 0),
                            flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
        self.value_ctrl = wx.TextCtrl(self, -1, size=(120, -1))
        name_ctrl_sizer.Add(self.value_ctrl, (1, 1),
                            flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
        self.error_ctrl = wx.TextCtrl(self, -1, size=(120, -1))
        name_ctrl_sizer.Add(self.error_ctrl, (1, 2),
                            flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
        self.add_button = wx.Button(self, id=wx.ID_ANY, label='Add')
        name_ctrl_sizer.Add(self.add_button, (1, 3),
                            flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
        sizer.Add(name_ctrl_sizer)
        self.Bind(wx.EVT_BUTTON, self.OnAdd, self.add_button)

        line = wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        sizer.Add(line, 0, wx.GROW | wx.RIGHT | wx.TOP, 5)

        self.listbox = MyHtmlListBox(self, -1, size=(-1, 150),
                                     style=wx.BORDER_SUNKEN)
        self.listbox.SetItemList(self.lines)
        sizer.Add(self.listbox, 1, wx.GROW | wx.ALL, 10)

        self.delete_button = wx.Button(self, id=wx.ID_ANY, label='Delete')
        sizer.Add(self.delete_button, 0, wx.CENTRE, 0)
        self.Bind(wx.EVT_BUTTON, self.OnDelete, self.delete_button)

        button_sizer = wx.StdDialogButtonSizer()
        okay_button = wx.Button(self, id=wx.ID_OK)
        # okay_button.SetDefault()
        button_sizer.AddButton(okay_button)
        button_sizer.AddButton(wx.Button(self, id=wx.ID_CANCEL))
        button_sizer.Realize()
        self.Bind(wx.EVT_BUTTON, self.OnApply, okay_button)

        line = wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        sizer.Add(line, 0, wx.GROW | wx.RIGHT | wx.TOP, 5)

        sizer.Add(button_sizer, 0, wx.ALIGN_RIGHT, 5)
        self.SetSizer(sizer)
        sizer.Fit(self)
        self.Layout()

    def OnApply(self, event):
        '''OnApply(self, event) --> None

        Callback for ok button click or apply button
        '''
        event.Skip()

    def OnAdd(self, event):
        '''OnAdd(self, event) --> None

        Callback for adding an entry
        '''
        sigma = self.error_ctrl.GetValue()
        if sigma.strip()=='':
            line = '%s.new_var(\'%s\', %s)'%(self.var_name,
                                             self.name_ctrl.GetValue(), self.value_ctrl.GetValue())
        else:
            line = '%s.new_sys_err(\'%s\', %s, %s)'%(self.var_name, self.name_ctrl.GetValue(),
                                                     self.value_ctrl.GetValue(), sigma)
        try:
            self.model.eval_in_model(line)
        except Exception as e:
            result = 'Could not evaluate the expression. The python error'+ \
                     'is: \n'+e.__repr__()
            ShowWarningDialog(self, result, 'Error in expression')
        else:
            self.lines.append(line)
            self.listbox.SetItemList(self.lines)

    def OnDelete(self, event):
        '''OnDelete(self, event) --> None

        Callback for deleting an entry
        '''
        result = 'Do you want to delete the expression?\n'+ \
                 'Remember to check if parameter is used elsewhere!'
        result = ShowQuestionDialog(self, result, 'Delete expression?')
        if result:
            self.lines.pop(self.listbox.GetSelection())
            self.listbox.SetItemList(self.lines)

    def GetLines(self):
        '''GetLines(self) --> uservars lines [list]

        Returns the list user variables.
        '''
        return self.lines


class SimulationExpressionDialog(wx.Dialog):
    '''A dialog to edit the Simulation expression
    '''
    model: Model

    def __init__(self, parent, model, instruments, sim_func, arguments, inst_name,
                 data_index):
        '''Creates a SimualtionExpressionDialog.

        model - a Model object.
        instruments - a dictionary of possible instruments
        arguments - the arguments to the simulation function, a list of strings.
        sim_func - a string of the simulation function name.
        inst_name - the name of the current instrument.
        data_index - an integer for the current data index.
        '''
        if not model.compiled:
            model.compile_script()

        self.model = model
        self.instruments = instruments
        self.available_sim_funcs = list(self.model.eval_in_model('model.SimulationFunctions.keys()'))
        self.data_index = data_index

        # Do the layout of the dialog
        wx.Dialog.__init__(self, parent, -1, 'Simulation editor')
        self.SetAutoLayout(True)

        # Find out the maximum number of arguments to the available sim_funcs
        max_val = -1
        self.sim_args = {}
        self.sim_defaults = {}
        for func in self.available_sim_funcs:
            doc = self.model.eval_in_model('model.SimulationFunctions'
                                           '["%s"].__doc__'%func)
            doc_lines = find_code_segment(doc, 'Parameters').splitlines()
            max_val = max(len(doc_lines), max_val)
            args = []
            defaults = []
            for line in doc_lines:
                items = line.lstrip().rstrip().split(' ')
                args.append(items[0])
                defaults.append(items[1].replace('data', 'd'))
            self.sim_args[func] = args
            self.sim_defaults[func] = defaults

        expressions = {'Instrument': inst_name}
        for arg_name, arg in zip(self.sim_args[sim_func], arguments):
            expressions[arg_name] = arg

        if max_val<0:
            raise ValueError('Wrongly formatted function docs for the simulation functions')

        gbs = wx.GridBagSizer(2, max_val)

        # Creating the column labels
        col_labels = ['Simulation', 'Instrument']
        [col_labels.append(arg) for arg in self.sim_args[sim_func] if not arg in col_labels]
        self.labels = []
        self.arg_controls = []
        for index in range(2+max_val):
            label = wx.StaticText(self, -1, '')
            gbs.Add(label, (0, index), flag=wx.ALIGN_LEFT, border=5)
            self.labels.append(label)
            # If the expression is not an instrument or simulation function
            if index>1:
                exp_ctrl = wx.TextCtrl(self, -1, size=(100, -1))
                gbs.Add(exp_ctrl, (1, index),
                        flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
                self.arg_controls.append(exp_ctrl)

        for item, label in zip(col_labels[:2], self.labels[:2]):
            label.SetLabel(item)
        # Creating the text boxes for the arguments
        # Setting the text in the column labels and text controls
        for item, label, arg_ctrl in zip(col_labels[2:],
                                         self.labels[2:],
                                         self.arg_controls):
            label.SetLabel(item)
            arg_ctrl.SetValue(expressions[item])
            arg_ctrl.SetEditable(True)

        for i in range(len(col_labels)-2, len(self.arg_controls)):
            self.arg_controls[i].SetEditable(False)
            # self.arg_controls[i].Show(False)
            # self.arg_controls[i].SetValue('NA')

        # Creating the controls
        # Simulation choice control
        self.sim_choice = wx.Choice(self, -1,
                                    choices=self.available_sim_funcs)
        self.Bind(wx.EVT_CHOICE, self.on_sim_change, self.sim_choice)
        self.sim_choice.SetSelection(self.available_sim_funcs.index(sim_func))
        gbs.Add(self.sim_choice, (1, 0),
                flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)

        # Instrument choice control
        self.inst_choice = wx.Choice(self, -1,
                                     choices=list(self.instruments.keys()))
        # self.Bind(wx.EVT_CHOICE, self.on_inst_change, self.inst_choice)
        self.inst_choice.SetSelection(list(self.instruments.keys()).index(expressions['Instrument']))
        gbs.Add(self.inst_choice, (1, 1),
                flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)

        button_sizer = wx.StdDialogButtonSizer()
        okay_button = wx.Button(self, wx.ID_OK)
        okay_button.SetDefault()
        button_sizer.AddButton(okay_button)
        button_sizer.AddButton(wx.Button(self, wx.ID_CANCEL))

        button_sizer.Realize()
        self.Bind(wx.EVT_BUTTON, self.on_ok_button, okay_button)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(gbs, 1, wx.GROW | wx.ALL, 10)
        line = wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        sizer.Add(line, 0, wx.GROW | wx.RIGHT | wx.TOP, 5)
        sizer.AddSpacer(5)

        sizer.Add(button_sizer, 0, wx.ALIGN_RIGHT, 5)
        sizer.AddSpacer(5)
        self.SetSizer(sizer)
        sizer.Fit(self)
        self.Layout()

    def on_sim_change(self, evt):
        '''Callback for changing the choice widget for the different simulations.
        '''
        new_sim = self.sim_choice.GetStringSelection()
        # Update the column labels
        new_labels = []
        for label, arg_name in zip(self.labels[2:], self.sim_args[new_sim]):
            new_labels.append(label.GetLabel()!=arg_name)
            label.SetLabel(arg_name)
        # Clear the remaining column labels
        for label in self.labels[len(self.sim_args[new_sim])+2:]:
            label.SetLabel('')

        # Update the text controls - if needed
        for i in range(len(self.sim_args[new_sim])):
            # if new_labels[i]:
            if True:
                self.arg_controls[i].SetValue(self.sim_defaults[new_sim][i])
            self.arg_controls[i].SetEditable(True)
            # self.arg_controls[i].Show(True)
        # Hide and clear the remaining text controls
        for ctrl in self.arg_controls[len(self.sim_args[new_sim]):]:
            ctrl.SetEditable(False)
            ctrl.SetValue('')

            # ctrl.Show(False)

    def on_ok_button(self, event):
        '''Callback for pressing the ok button in the dialog'''
        expressions = self.GetExpressions()
        # Hack to get it working with d = data[0]
        exec('d = data[%d]'%self.data_index, self.model.script_module.__dict__)
        for exp in expressions:
            try:
                self.model.eval_in_model(exp)
            except Exception as e:
                result = ('Could not evaluate expression:\n%s.\n'%exp+
                          ' The python error is: \n'+e.__repr__())
                ShowWarningDialog(self, result, 'Error in expression')
            else:
                event.Skip()

    def GetExpressions(self):
        ''' Returns the current expressions in the dialog box '''
        return [ctrl.GetValue() for ctrl in self.arg_controls
                if ctrl.IsEditable()]

    def GetInstrument(self):
        ''' Returns the selected instrument, a string'''
        return self.inst_choice.GetStringSelection()

    def GetSim(self):
        ''' Returns the selected simulation, a string'''
        return self.sim_choice.GetStringSelection()


class ParameterExpressionDialog(wx.Dialog):
    ''' A dialog for setting parameters for fitting
    '''
    model: Model

    def __init__(self, parent, model, expression=None, sim_func=None):
        wx.Dialog.__init__(self, parent, -1, 'Parameter editor')
        self.SetAutoLayout(True)
        self.model = model
        self.sim_func = sim_func
        if not model.compiled:
            model.compile_script()

        gbs = wx.GridBagSizer(2, 3)

        col_labels = ['Object', 'Parameter', 'Expression']

        for item, index in zip(col_labels, list(range(len(col_labels)))):
            label = wx.StaticText(self, -1, item)
            gbs.Add(label, (0, index), flag=wx.ALIGN_LEFT, border=5)

        # Get the objects that should be in the choiceboxes
        par_dict = model.get_possible_parameters()
        objlist = []
        funclist = []
        for cl in par_dict:
            obj_dict = par_dict[cl]
            for obj in obj_dict:
                objlist.append(obj)
                funclist.append(obj_dict[obj])

        self.objlist = objlist
        self.funclist = funclist
        self.obj_choice = wx.Choice(self, -1, choices=objlist)
        self.Bind(wx.EVT_CHOICE, self.on_obj_change, self.obj_choice)

        self.func_choice = wx.Choice(self, -1)
        # This will init func_choice
        self.obj_choice.SetSelection(0)

        gbs.Add(self.obj_choice, (1, 0),
                flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
        gbs.Add(self.func_choice, (1, 1),
                flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)

        exp_right = ''
        if expression:
            p = expression.find('(')
            exp_left = expression[:p]
            obj = exp_left.split('.')[0]
            func = exp_left.split('.')[1]
            exp_right = expression[p+1:-1]
            obj_pos = [i for i in range(len(objlist)) if objlist[i]==obj]
            if len(obj_pos)>0:
                self.obj_choice.SetSelection(obj_pos[0])
                self.on_obj_change(None)
                func_pos = [i for i in range(len(funclist[obj_pos[0]])) \
                            if funclist[obj_pos[0]][i]==func]
                if len(func_pos)>0:
                    self.func_choice.SetSelection(func_pos[0])
                else:
                    raise ValueError('The function %s for object %s does not exist'%(func, obj))
            else:
                raise ValueError('The object %s does not exist'%obj)

        # self.expression_ctrl = wx.TextCtrl(self, -1, exp_right,\
        #                       size=(300, -1))

        self.expression_ctrl = ParameterExpressionCombo(par_dict, sim_func, self, -1, exp_right,
                                                        size=(300, -1))
        gbs.Add(self.expression_ctrl, (1, 2),
                flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)

        button_sizer = wx.StdDialogButtonSizer()
        okay_button = wx.Button(self, wx.ID_OK)
        okay_button.SetDefault()
        button_sizer.AddButton(okay_button)
        button_sizer.AddButton(wx.Button(self, wx.ID_CANCEL))
        # apply_button = wx.Button(self, wx.ID_APPLY)
        # apply_button.SetDefault()
        # button_sizer.AddButton(apply_button)
        button_sizer.Realize()
        self.Bind(wx.EVT_BUTTON, self.OnApply, okay_button)
        # self.Bind(wx.EVT_BUTTON, self.OnApply, apply_button)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(gbs, 1, wx.GROW | wx.ALL, 10)
        line = wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        sizer.Add(line, 0, wx.GROW | wx.RIGHT | wx.TOP, 5)
        sizer.AddSpacer(5)

        sizer.Add(button_sizer, 0, wx.ALIGN_RIGHT, 5)
        sizer.AddSpacer(5)
        self.SetSizer(sizer)
        sizer.Fit(self)
        self.Layout()

    def on_obj_change(self, event):
        '''on_obj_change(self, event) --> None

        On changing the object the funclist should be updated
        '''
        index = self.obj_choice.GetSelection()
        self.func_choice.SetItems(self.funclist[index])

    def OnApply(self, event):
        '''OnApply(self, event) --> None
        '''
        evalstring = self.GetExpression()
        try:
            self.model.eval_in_model(evalstring)
        except Exception as e:
            result = 'Could not evaluate the expression. The python'+ \
                     'is: \n'+e.__repr__()
            ShowWarningDialog(self, result, 'Error in expression')
        else:
            event.Skip()

    def GetExpression(self):
        '''GetExpression(self) --> expression

        Yields the string that has been edited in the dialog
        '''
        objstr = self.obj_choice.GetStringSelection()
        funcstr = self.func_choice.GetStringSelection()
        set_expression = self.expression_ctrl.GetValue()
        evalstring = '%s.%s(%s)'%(objstr, funcstr, set_expression)

        return evalstring




'''
==========================================
:mod:`orso` ORSO standard file data loader
==========================================

Loads the format following the specification of the Open Reflectometry Standards Organization (ORSO).
See https://www.reflectometry.org/working_groups/file_formats/ for more details
'''

import numpy as np
import re

from typing import List
from ..data_loader_framework import Template
from ..utils import ShowWarningDialog
from orsopy.fileio import load_orso, OrsoDataset

class Plugin(Template):
    wildcard='*.ort'
    _cached_file = ''
    _cached_data: List[OrsoDataset] = None

    def __init__(self, parent):
        Template.__init__(self, parent)
        self.x_col=0
        self.y_col=1
        self.e_col=3
        self.xe_col=2
        self.comment='#'
        self.skip_rows=0
        self.delimiter=None

    def CanOpen(self, file_path):
        if not Template.CanOpen(self, file_path):
            return False
        l1=open(file_path, 'r', encoding='utf-8').readline()
        return l1.startswith('# # ORSO')

    def LoadCached(self, file_path):
        if self._cached_file==file_path:
            return self._cached_data
        else:
            orso_datasets = load_orso(file_path)
            self._cached_file=file_path
            self._cached_data=orso_datasets
            return orso_datasets

    def CountDatasets(self, file_path):
        try:
            orso_datasets=self.LoadCached(file_path)
        except Exception as e:
            return 1
        else:
            return len(orso_datasets)

    def LoadData(self, dataset, filename, data_id=0):
        '''
        Loads the data from filename into the data_item_number.
        '''
        try:
            orso_datasets=self.LoadCached(filename)
        except Exception as e:
            ShowWarningDialog(self.parent, 'Could not load the file: '+filename+
                              ' \nPlease check the format.\n\n ORSO reader gave the following error:\n'+str(e))
            return
        else:
            orso_dataset: OrsoDataset=orso_datasets[data_id]
            data: np.ndarray=orso_dataset.data.T

            cols=orso_dataset.info.columns

            start_usercols=3
            dataset.x_raw=np.asarray(data[0])
            dataset.y_raw=np.asarray(data[1])
            dataset.error_raw=np.asarray(data[2])
            if data.shape[0] > 3 and getattr(cols[3], 'error_of')==cols[0].name:
                start_usercols=4
                dataset.set_extra_data('res', np.asarray(data[3])*cols[3].to_sigma, 'res')
            for i, col in enumerate(data[start_usercols:]):
                colname=cols[i+start_usercols].name
                if not colname.isidentifier() or colname in ['lambda', 'x', 'y', 'error', 'res']:
                    colname='col_'+re.sub('[^0-9a-zA-Z_]', '', colname)
                dataset.set_extra_data(colname, np.asarray(col), colname)
            # Name the dataset accordign to file name
            if orso_dataset.info.data_set:
                dataset.name=str(orso_dataset.info.data_set)
            # Run the commands on the data - this also sets the x,y, error memebers
            # of that data item.
            dataset.run_command()

            # insert metadata into ORSO compatible fields
            dataset.meta=orso_dataset.info.to_dict()
            if 'data_set' in dataset.meta:
                del(dataset.meta['data_set'])



'''
==================================================
:mod:`amor` Amor at SINQ reflectometer data loader
==================================================

Loads the default datafile format used for extracted reflectivity at
the SINQ reflectometer Amor.
'''

import numpy as np
import os

from ..data_loader_framework import Template
from ..utils import ShowWarningDialog

class Plugin(Template):
    wildcard='*.Rqz'

    def __init__(self, parent):
        Template.__init__(self, parent)
        self.x_col=0
        self.y_col=1
        self.e_col=3
        self.xe_col=2
        self.comment='#'
        self.skip_rows=0
        self.delimiter=None

    def LoadData(self, dataset, filename, data_id=0):
        '''LoadData(self, data_item_number, filename) --> none

        Loads the data from filename into the data_item_number.
        '''
        # get scan number and polarization channel from filename
        name=os.path.basename(filename)[:-4]
        fhandle=open(filename, 'r')
        fline=fhandle.readline()
        while not 'q_z' in fline:
            if 'Input file indices:' in fline:
                name+=fline.split(':')[1].strip()
            if 'Extracted states:' in fline:
                name+=' (%s)'%(fline.split(':')[1].strip().split(' ')[0])
            fline=fhandle.readline()
            if not fline:
                ShowWarningDialog(self.parent, 'Could not load the file: '+filename+' \nWrong format.\n')
                return

        try:
            load_array=np.loadtxt(fhandle, delimiter=self.delimiter, comments=self.comment, skiprows=self.skip_rows)
        except Exception as e:
            ShowWarningDialog(self.parent, 'Could not load the file: '+filename+
                              ' \nPlease check the format.\n\n numpy.loadtxt'+' gave the following error:\n'+str(e))
            return
        else:
            # For the freak case of only one data point
            if len(load_array.shape)<2:
                load_array=np.array([load_array])
            # Check so we have enough columns
            if load_array.shape[1]-1<max(self.x_col, self.y_col, self.e_col):
                ShowWarningDialog(self.parent, 'The data file does not contain'+'enough number of columns. It has '
                                  +str(load_array[1])+' columns. Rember that the column index start at zero!')
                # Okay now we have showed a dialog lets bail out ...
                return
            # The data is set by the default Template.__init__ function, neat hu
            # Know the loaded data goes into *_raw so that they are not
            # changed by the transforms
            dataset.x_raw=load_array[:, self.x_col]
            dataset.y_raw=load_array[:, self.y_col]
            dataset.error_raw=load_array[:, self.e_col]
            dataset.set_extra_data('res', load_array[:, self.xe_col], 'res')
            # Name the dataset accordign to file name
            dataset.name=name
            # Run the commands on the data - this also sets the x,y, error memebers
            # of that data item.
            dataset.run_command()

            # insert metadata into ORSO compatible fields
            dataset.meta['data_source']['facility']='SINQ@PSI'
            dataset.meta['data_source']['experiment']['instrument']='Amor'
            dataset.meta['data_source']['experiment']['probe']='neutron'
            dataset.meta['data_source']['measurement']['scheme']='angle- and energy-dispersive'
            dataset.meta['reduction']={'software': {'name': 'eos/amor-reducer'}}
